import csv
from bytegain.zap.utils import gcs
import logging
from google.cloud import storage
import tempfile
import json
import dateutil.parser as dt_parser

logger = logging.getLogger('combine_al_predictions')


def run(driver_input, map_config, stages, dataflow_args, input_predictions, output_predictions, tmp_dir):
    """
    Reads AL BG user value model predictions, which are obtained via running driver in predict mode, and
    generates csv with columns (userId,signup_timestamp,prediction,current_lease_status)
    and another csv with columns (date,sum_predictions,average_prediction)
    """
    if not input_predictions.endswith('/'):
        input_predictions += '/'

    if not output_predictions.endswith('/'):
        output_predictions += '/'

    # Download prediction files
    bucket, gs_path = gcs.parse_gs_url(input_predictions)

    client = storage.Client()
    bucket = client.bucket(bucket)
    files = []
    blobs = bucket.list_blobs(prefix=gs_path)
    for blob in blobs:
        filename = blob.name.rsplit('/', 1)[1]
        if filename.startswith('prediction.results'):
            files.append(filename)
    local_files = []
    local_tmp_dir = tempfile.mkdtemp() + '/'
    for filename in files:
        blob = bucket.blob(gs_path + filename)
        local_path = local_tmp_dir + filename
        logging.info("File downloaded: %s" % local_path)
        blob.download_to_filename(local_path)
        local_files.append(local_path)

    # Get users and probabilities
    users = []
    for local_file in local_files:
        with open(local_file, 'r') as input_file:
            logging.info("File opened: %s" % local_file)
            for row in input_file:
                row_json = json.loads(row)
                user_id, signup_timestamp = row_json['id'].split('_')
                prediction = row_json['probability'][0]
                label = row_json['label']
                users.append((user_id, signup_timestamp, prediction, label))

    assert len(users) > 0

    # Sort by signup timestamp
    logger.info("Sorting by signup timestamp")
    users = sorted(users, key=lambda user: user[1])

    logger.info("Read %d users" % len(users))

    def extract_date(timestamp):
        return dt_parser.parse(timestamp)

    def extract_day(timestamp):
        return timestamp[8:10]

    # Go through sorted users and generate daily stats
    logger.info("Generating stats")
    stats = []
    preds_sum = 0.0
    user_count = 0
    prev_day = extract_day(users[0][1])
    for user in users:
        cur_day = extract_day(user[1])
        if cur_day != prev_day:
            date = user[1][:10]
            aver_pred = preds_sum / user_count
            logging.info('Sum of predictions: %.4f' % preds_sum)
            logging.info('Average prediction: %.4f' % aver_pred)
            stats.append((date, preds_sum, aver_pred))
            preds_sum = 0.0
            user_count = 0
        preds_sum += float(user[2])
        user_count += 1
        prev_day = cur_day

    start_date = stages['sessions']['data_start_date']
    end_date = stages['sessions']['data_end_date']

    predictions_output_filename = 'AL_user_value_predictions_%s_%s.csv' % (start_date, end_date)
    predictions_output_file = local_tmp_dir + predictions_output_filename

    with open(predictions_output_file, "w") as o_file:
        writer = csv.writer(o_file)
        writer.writerow(["userId", "signup_timestamp", "prediction", "current_lease_status"])
        writer.writerows(users)

    logger.info("Local file with predictions: %s" % predictions_output_file)

    stats_output_filename = 'AL_user_value_stats_%s_%s.csv' % (start_date, end_date)
    stats_output_file = local_tmp_dir + stats_output_filename

    with open(stats_output_file, "w") as o_file:
        writer = csv.writer(o_file)
        writer.writerow(["date", "sum_predictions", "average_prediction"])
        writer.writerows(stats)

    logger.info("Local file with stats: %s" % stats_output_file)

    logger.info("Cloud output dir: %s" % output_predictions)

    gcs.upload_local_files(
        local_file_paths=[predictions_output_file, stats_output_file],
        gs_path=output_predictions
    )

    logger.info("Cloud output dir: %s" % output_predictions)

    return {'n_users': len(users)}



import psycopg2
import time
import sys
from bytegain.custom.data_import.schema_info import REQUIRED_COLUMNS, ColumnInfo

class ConnectionFactory(object):
    def __init__(self, db_url):
        self._conn = None
        self._db_url = db_url

    def get_connection(self):
        try:
            # This is a trick to trigger error if connection is timed out
            if self._conn is not None and self._conn.isolation_level is not None:
                return self._conn
        except psycopg2.OperationalError as oe:
            print ("DB timeout: reconnecting")

        self._conn = psycopg2.connect(self._db_url)
        self._conn.autocommit = True
        # TODO - how to generalize this?
        self._conn.cursor().execute('set query_group TO "slow"')

        return self._conn

def create_table(cur, table_info, sortkey):
    columns = []
    for column in table_info.columns:
        columns.append("%s %s" % (column.name, column.datatype))

    drop = "drop table if exists %s" % (table_info.full_name)
    cur.execute(drop)
    create = "create table %s (%s) distkey(%s) sortkey(%s)" % (
            table_info.full_name, ", ".join(columns), sortkey, sortkey
    )
    print(create)
    cur.execute(create)

def copy_to_s3_from_redshift(cur, select, table_info, region, s3_secret_key, name):
    return copy_to_s3_from_redshift_by_bucket(cur, select, table_info, get_s3_bucket(region), s3_secret_key, name)

def copy_to_s3_from_redshift_by_bucket(cur, select, table_info, s3_bucket, s3_secret_key, name, destination_folder=""):
    start_time = time.time()
    statement = select.replace("'", "\\'")
    dest_csv = "%s/%s_%s/%s." % (destination_folder, table_info.schema, table_info.table, name)
    manifest = dest_csv + "manifest"
    statement = """
    unload('%s') to 's3://%s/%s'
    credentials 'aws_access_key_id=AKIAIGHWWWWWV534ALTA;aws_secret_access_key=%s'
    ALLOWOVERWRITE PARALLEL ON GZIP ESCAPE MANIFEST
    """ % (statement, s3_bucket, dest_csv, s3_secret_key)

    print("statement: %s" % statement)
    cur.execute(statement)
    sys.stdout.write(("Unloaded %s in %ds. " % (name, time.time() - start_time)))
    sys.stdout.flush()
    return manifest

def copy_from_s3(cur, table_info, manifest, region):
    manifest_url = 's3://%s/%s' % (get_s3_bucket(region), manifest)
    credentials = 'aws_iam_role=arn:aws:iam::087456767550:role/redshift-data-loader'
    upload_statement = "copy %s.%s from '%s' credentials '%s' MANIFEST ESCAPE GZIP COMPUPDATE ON EMPTYASNULL REGION '%s'" % (
            table_info.schema, table_info.table, manifest_url, credentials, region)
    upload_start = time.time()
    cur.execute(upload_statement)
    sys.stdout.write(("Uploaded %s.%s in: %ds.  " % (table_info.schema, table_info.table, time.time() - upload_start)))
    sys.stdout.flush()

def get_segment_tables(connection_factory, schema):
    conn = connection_factory.get_connection()
    cur = conn.cursor();
    cur.execute("select tablename from pg_tables where schemaname = '%s'" % schema)
    tables = []
    for row in cur:
        table = row[0]
        if table == "tracks":
            continue
        else:
            tables.append(table)

    return tables

def compute_table_info(connection_factory, table_info, where, exclude_common=None, no_groups=False,
                       force_required_columns = True, included_columns = None):
    """ Get ColumnInfos for all columns in table.
    """
    cur = connection_factory.get_connection().cursor()
    columns = get_meta_data(cur, table_info, True)
    if force_required_columns:
        for column in REQUIRED_COLUMNS:
            if column not in [c[0] for c in columns]:
                print("%s missing required column %s" % (table_info, column))
                return None

    excluded = exclude_common.columns if exclude_common else []
    excluded_set = set([x.name for x in excluded if x.name not in REQUIRED_COLUMNS])
    columns = [x for x in columns if x[0] not in excluded_set]
    if included_columns:
        new_columns = []
        for columnname in included_columns:
            for column in columns:
                if column[0] == columnname:
                    new_columns.append(column)
                    break
        columns = new_columns

    if len(columns) == 0:
        print("%s had no non-duplicate columns" % table_info)
        return table_info

    column_selects = []
    for column in columns:
        info = ColumnInfo(column[0], column[1])
        column_selects.extend(info.select_stats())
        table_info.add_column(info)

    if no_groups:
        return table_info

    sys.stdout.write("Counting %s                         \r" % table_info.table)
    sys.stdout.flush()
    where = table_info.get_full_where(where)
    select = "select count(1), %s from %s %s" % (",".join(column_selects), table_info.table, where)
    try:
        cur.execute(select)
    except psycopg2.ProgrammingError as e:
        print(e)
        print("error accessing table %s" % (table_info))
        return None
    row = cur.fetchone()
    items = iter(row)
    rows = next(items)
    table_info.set_rows(rows)
    for column_info in table_info.columns:
        column_info.read_stats(items)

    for column_info in table_info.columns:
        if column_info.is_numeric():
            select = 'select median("%s"), PERCENTILE_CONT ( 0.25 ) WITHIN GROUP (ORDER BY "%s"), '\
            'PERCENTILE_CONT ( 0.75 ) WITHIN GROUP (ORDER BY "%s")'\
            ' from %s %s' % (column_info.name, column_info.name, column_info.name, table_info.table, where)
            cur.execute(select)
            row = cur.fetchone()
            if row[0] is None:
                print("No median values for: %s " % column_info.name)
            else:
                column_info.median = float(row[0])
                column_info.percentile_25 = float(row[1])
                column_info.percentile_75 = float(row[2])
        else:
            print("%s is not numeric: %s" % (column_info.name, column_info.datatype))

    for column_info in table_info.columns:
        sys.stdout.write("Grouping %s.%s             \r" % (table_info.table, column_info.name))
        sys.stdout.flush()
        select = "select \"%s\", count(*) from %s %s group by 1 order by 2 desc limit 200" % (column_info.name, table_info.table, where)
        cur.execute(select)
        for row in cur:
            outcome = str(row[0], "UTF-8") if type(row[0]) == str else row[0]
            count = row[1]
            fraction = float(count) / table_info.row_count
            column_info.add_group(outcome, fraction)

    return table_info

def get_meta_data(cur, table_info, include_datatype = True):
    """ Get column meta data for table from DB cursor.
    """
    cur.execute("set search_path to '%s'" % table_info.schema)
    columns = []
    select = "select \"column\", \"type\" from pg_table_def where tablename = '%s'" % table_info.table
    cur.execute(select)
    # Ordering is important, so don't use hash/set
    for row in cur.fetchall():
        columns.append((row[0], row[1]) if include_datatype else row[0])
    return columns

def create_date_clause(start_date, end_date, tablename = None):
    # Create Where clause from start/end dates
    if tablename is None:
        column = "received_at"
    else:
        column = "%s.received_at" % tablename

    where = ""
    if start_date:
        where = "%s > timestamp '%s'" % (column, start_date)
    if end_date:
        if len(where) > 0:
            where = where + " AND "
        where = where + "%s < timestamp '%s'" % (column, end_date)
    return where

def and_where(w1, w2):
    if w1 and w2:
        return '(%s) AND (%s)' % (w1, w2)
    elif w1:
        return w1
    elif w2:
        return w2
    else:
        return None



def pages_screens_compass(rows):
    """
    Keep pages, screens and compass events
    :param rows:
    :return:
    """
    rows = [row for row in rows
            if (
                row.get('type', '') == 'page'
                or row.get('type', '') == 'screen'
                or row.get('type', '') == 'compass'
            )
            ]
    return rows


def pages_screens(rows):
    """
    Keep pages and screens events
    :param rows:
    :return:
    """
    rows = [row for row in rows
            if (
                row.get('type', '') == 'page'
                or row.get('type', '') == 'screen'
            )
            ]
    return rows



import pyximport

pyximport.install()

import bytegain.custom.cust.apartmentlist.al_xgb as al_xgb
import bytegain.nn_tools.auc_plot as auc_plot
import bytegain.custom.cust.apartmentlist.leads_filter_160 as lf160
import time

import bytegain.custom.nn_tools.xg_classifier_test as test_gxb

# import leads_from_redshift as tfl
# feature_file = 'features/al_1.6.1.json'
# features = tfl.run_feature_analysis(feature_file)

feature_file = 'features/al_1.6.1.json'
rh, datasets = al_xgb.create_rh_and_dataset(feature_file, downsample_negative=False, random_seed=2828712, filter = lf160.filter)

n_rounds = 500
# XGBoost classifier
print("Training the default XGBoost classifier")
time0 = time.time()
classifier = test_gxb.XgbClassifier()
classifier.train(datasets, max_rounds=n_rounds, max_depth=4, l2_norm=1000, min_child_weight=200)
print("Feature importances")
feature_importances = classifier._classifier.feature_importances_.tolist()
imp_list = []
for i in range(len(feature_importances)):
    imp_list.append((rh._buckets[i]._name, feature_importances[i]))
    # print "Index %d, %s: %s"%(i, rh._buckets[i]._name, feature_importances[i])
imp_list.sort(key=lambda x: x[1], reverse=True)
for i in range(len(imp_list)):
    print("%25s: %4.1f"%(imp_list[i][0], imp_list[i][1]*100))
results = classifier.get_sorted_results(datasets.test)
# results = test_gxb.cv(datasets, xgb.XgbClassifier, 500, parallel=True, max_depth=4, l2_norm=1000, train_step_size=0.3, subsample=0.9, min_child_weight=50)
print("--------------------")
print("The AUC for the default XGBoost classifier is %.5f"%auc_plot.compute_auc(results))
print("--------------------")
auc_plot.plot_auc('auc_xgb.png', results)
# Feature selection
# classifier.model_select_features(datasets, 500)

# ===================================================
# Default XGBooster
print("Training the defaut XGBooster")
time0 = time.time()
classifier = test_gxb.XgbClassifier(model='booster')
classifier.train(datasets, max_rounds=500, max_depth=4, l2_norm=1000, min_child_weight=200)
print("Took %.5f sec to train the defaut XGBooster"%(time.time() - time0))
features = classifier.feature_scores2(rh)
print("--------------------")
for feature in features:
    print("%25s: %4.1f" % (feature[0], feature[1]))
results = classifier.get_sorted_results(datasets.test)
print("The AUC for the default booster is %.5f"%auc_plot.compute_auc(results))
auc_plot.plot_auc('auc_xgb.png', results)

# ===================================================
# Tuned XGBooster
print("Training the tuned XGBooster")
time0 = time.time()
classifier = test_gxb.XgbClassifier(model='booster_tuned')
classifier.train(datasets, max_rounds=n_rounds, max_depth=4, l2_norm=1000, min_child_weight=200)
print("Took %.5f sec to train the tuned XGBooster"%(time.time() - time0))
features = classifier.feature_scores2(rh)
print("--------------------")
for feature in features:
    print("%25s: %4.1f" % (feature[0], feature[1]))
results = classifier.get_sorted_results(datasets.test)
# results = test_gxb.cv(datasets, xgb.XgbClassifier, 401, parallel=True, max_depth=4, l2_norm=1000, train_step_size=0.3, subsample=0.9, min_child_weight=50)
print("--------------------")
print("The AUC for the tuned booster is %.5f"%auc_plot.compute_auc(results))
print("--------------------")
auc_plot.plot_auc('auc_xgb_tuned.png', results)



import psycopg2
import argparse
import time
import sys
import os
import tempfile
import gzip
from threading import Thread
from queue import Queue
import datetime
import io


parser = argparse.ArgumentParser(description='Joins a set of segment event tables together with main table')
parser.add_argument('--db', type=str, help='host:port:db:user:pass:schema:table', required=True)
parser.add_argument('--output_table', type=str, help='Table to create', required=True)
parser.add_argument('--input_tables', type=str, help='Comma-separated list of input tables', required=True)
parser.add_argument('--output_sortkeys', type=str, help='Sort/dist keys for output table', required=True)
parser.add_argument('--main_table', type=str, help='tracks table', required=True)
args = parser.parse_args()

src_conn, src_cur, schema, src_table = parse_params(args.db)

def parse_params(params):
    s = params.split(':')
    assert len(s) == 7
    conn = psycopg2.connect(host=s[0], port=s[1], dbname=s[2], user=s[3], password=s[4], sslmode='require')
    conn.autocommit = True
    cur = conn.cursor()
    if s[5]:
        cur.execute("set search_path to '%s'" % s[5])
    return conn, cur, s[5], s[6]

def get_meta_data(table):
    columns = []
    src_cur.execute("select \"column\" from pg_table_def where tablename = '%s'" % table)
    # Ordering is important, so don't use hash/set
    for row in src_cur.fetchall():
        name = row[0]
        columns.append(name)
    return columns

def get_columns():
    start_time = time.time()
    main_columns = get_meta_data(args.main_table)
    join_columns = {}

    for table in args.input_tables.split(','):
        columns = get_meta_data(table)
        included_columns = []
        for column in columns:
            if column not in main_columns and not column.startswith("context_"):
                if column in join_columns:
                    join_columns[column].append(table)
                else:
                    join_columns[column] = [table]

    return main_columns, join_columns

def create_joined_table(main_columns, join_columns):
    columns = []
    for column in main_columns:
        columns.append("%s.%s\n" % (args.main_table, column))

    for column in join_columns:
        tables = join_columns[column]
        if len(tables) == 1:
            columns.append(column)
        else:
            case = "case "
            for table in tables:
                case += "\twhen %s.%s is not null then %s.%s \n" % (table, column, table, column)
            case += "\telse NULL end as %s\n" % column
            columns.append(case)

    from_clause = "%s.%s\n" % (schema, args.main_table)
    for table in args.input_tables.split(","):
        from_clause += "left outer join %s.%s using (id)\n" % (schema, table)

    select = "select %s from %s" % (",".join(columns), from_clause)
    create = "create table %s as (%s) sortkey(%s) distkey(%s)" % (args.output_table, select, args.output_sortkeys, args.output_sortkeys)

    print("Create: %s" % create)
    src_cur.execute(create)

main_columns, join_columns = get_columns()
print(main_columns)
print(join_columns)
create_joined_table(main_columns, join_columns)



EXCLUDED_FEATURES = [
    'billable_lease_at_property',
    'billable_lease_at_property_180',
    'billable_lease_at_property_90',
    'data_updated_at',
    'leased_at_property',
    'leased_at_property_180',
    'leased_at_property_90',
    'lsd_control',
    'pct_leases_imported',
    'rental_id',
    'user_id',
    'value_at',
    'value_on',
]



import abc
import csv
import json
import os
import os.path
import tempfile

from bytegain.custom.iox.file.wofs import Wofs

GCS_BUCKET = 's3://al-bytegain'
GCS_PROJECT = 'bg-rest-api'

class BaseModelSpec(object, metaclass=abc.ABCMeta):

    def __init__(self, model_name, path):
        assert '/' not in model_name, "model_name may not contain '/'"
        self._model_name = model_name
        self._path = path
        if self._path is None:
            self._path = tempfile.mkdtemp("model")

        if len(self._path) > 0 and self._path[-1] != '/':
            self._path += '/'

    @property
    def model_name(self):
        return self._model_name

    @abc.abstractmethod
    def get_all_files(self, for_prod):
        pass

    def _create_name(self, extension):
        return "%s%s.%s" % (self._path, self._model_name, extension)

    def get_model_file(self):
        return self._create_name("model")

    def get_row_handler_file(self):
        return self._create_name("rh")

    def get_decile_file(self):
        return self._create_name("decile")

    def get_validation_file(self):
        return self._create_name("v.csv")

    def get_json_file(self):
        return self._create_name("json")

    def save(self, model, row_handler, results):
        model.save(self.get_model_file())
        row_handler.save(self.get_row_handler_file())
        save_deciles(self.get_decile_file(), results)
        save_results_as_csv(self.get_validation_file(), "id", "pred", results)

    def download(self, dir_path=None, for_prod=True):
        """
        Download model files from dir_path into the local files referenced this ModelSpec
        :param dir_path: The dir containing the files to download.  If None, defaults to ``GCS_BUCKET/self.model_name``.
        :param for_prod:
        :return: self
        """
        if dir_path is None:
            dir_path = GCS_BUCKET + '/' + self.model_name
        wofs = Wofs(
            aws_access_key_id=os.environ["AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=os.environ["AWS_SECRET_ACCESS_KEY"]
        )
        for filepath in self.get_all_files(for_prod=for_prod):
            basename = os.path.basename(filepath)
            wofs.download(os.path.join(dir_path, basename), filepath)

    def upload(self, dir_path=None):
        # type: (str) -> None
        """
        Copy the model files into dir_path.  For dir_path in S3, profile 'model-writer' is used.
        :param dir_path: A dirname path in the syntax of bytegain.io.file.wofs, e.g.,
         s3://bucket/k/e/y or /tmp/junk.  If None, defaults to ``GCS_BUCKET/self.model_name``.
        :return: None
        """
        if dir_path is None:
            dir_path = GCS_BUCKET + '/' + self.model_name
        wofs = Wofs(
            aws_access_key_id=os.environ["AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=os.environ["AWS_SECRET_ACCESS_KEY"]
        )
        for filepath in self.get_all_files(for_prod=False):
            basename = os.path.basename(filepath)
            wofs.upload(filepath, os.path.join(dir_path, basename))


class ModelSpec(BaseModelSpec):
    def __init__(self, model_name, path):
        super(ModelSpec, self).__init__(model_name, path)

    def get_model_meta_file(self):
        return self._create_name("model.meta")

    def get_all_files(self, for_prod=True):
        files = [self.get_model_file(), self.get_model_meta_file(), self.get_row_handler_file(), self.get_decile_file()]
        if not for_prod:
            files.append(self.get_validation_file())
        return files


class XgbModelSpec(BaseModelSpec):
    def __init__(self, model_name, path):
        super(XgbModelSpec, self).__init__(model_name, path)

    def get_all_files(self, for_prod=True):
        files = [self.get_model_file(), self.get_row_handler_file(), self.get_decile_file()]
        if not for_prod:
            files.append(self.get_validation_file())
        return files


def save_deciles(filename, results):
    vals = []
    results.reverse()
    for result in results:
        vals.append(result['probability'])
    length = len(vals)
    with open(filename, 'w') as output:
        output.write("0.0, ")
        for i in range(1, 100):
            val = vals[int(length * i / 100)]
            output.write("%6.4f," % val)
        output.write("1.0\n")


def save_results_as_csv(filename, id_header, prediction_header, results):
    with open(filename, "w") as f:
        out = csv.writer(f)
        out.writerow(['user_id', 'rental_id', 'predicted_score'])
        for result in results:
            user_id, property_id = result['id'].split('p')
            out.writerow([user_id, 'p%s' % property_id, result['probability']])


def save_json_spec(filename, model_name, row_handler):
    feature_spec = row_handler.get_feature_spec()
    feature_spec["model"] = model_name
    with open(filename, 'w') as output:
        json.dump(feature_spec, fp=output, sort_keys=True, indent=4)



import argparse
import os
from collections import defaultdict
import time
from bytegain.custom.data_import.clone_redshift_table import TableCloner
from bytegain.custom.data_import.schema_info import read_schema_from_file, save_schema, TableInfo, ColumnInfo

import bytegain.data_import.segment_utils as segment_utils
import bytegain.custom.data_import.db_utils as db_utils

def comma_list(s):
    if s:
        return s.split(",")
    return None

def create_joined_table(cur, table_info, select, sortkey):
    cur.execute("drop table if exists %s" % table_info.full_name)

    statement = "create table %s distkey(%s) sortkey(%s) as (%s)" % (
            table_info.full_name, sortkey, sortkey, select)
    print(statement)
    cur.execute(statement)

def get_join_select(schema_info, dst_schema, output_table, where):
    """ Creates a join select statement.  Returns select statement and
    a TableInfo representing the output of the Select.
    """
    start_time = time.time()
    columns = []
    src_columns = defaultdict(list)
    src_columns_table = defaultdict(list)
    src_columns_datatype = defaultdict(list)
    main_table = schema_info.main_table
    src_schema = schema_info.schema
    out_table_info = TableInfo(dst_schema, output_table)

    for column in main_table.columns:
        out_table_info.add_column(ColumnInfo(column.name, column.datatype))
        columns.append("%s.%s.%s\n" % (src_schema, main_table.table, column.name))

    for src_table in list(schema_info.join_tables.values()):
        for column in src_table.columns:
            if not main_table.has_column(column.name):
                src_columns[column.name].append(column)
                src_columns_table[column.name].append(src_table.table)
                src_columns_datatype[column.name].append(column.datatype)

    for columnname in src_columns:
        tables = src_columns_table[columnname]
        if len(tables) == 1:
            out_table_info.add_column(ColumnInfo(columnname, src_columns[columnname][0].datatype))
            columns.append("%s.%s" % (tables[0], columnname))
        else:
            any_char = False
            datatypes = src_columns_datatype[columnname]
            datatype = datatypes[0]
            # Char len varies, but ignore that
            all_same = all(x==datatypes[0] for x in datatypes) or all('char' in x for x in datatypes)
            if not all_same:
                print("Some varying datatypes for %s: %s" % (columnname, str(datatypes)))
                any_char = any('char' in x for x in datatypes)
                any_bool = any('bool' in x for x in datatypes)
            case = "case "
            for table, datatype in zip(tables, datatypes):
                value = "%s.%s" % (table, columnname)
                if not all_same:
                    if any_char:
                        if 'char' not in datatype:
                            # convert to char
                            print("Converting %s to char" % value)
                            if 'bool' in datatype:
                                #Cannot convert bool directly to char
                                value = "%s::int::char" % value
                            else:
                                value = "%s::char" % value
                            datatype = "varchar"
                    elif any_bool:
                        if 'bool' in datatype:
                            print("Converting %s to int" % value)
                            value = "%s::int" % value
                            datatype = "int"
                case += "\twhen %s.%s is not null then %s \n" % (table, columnname, value)
            case += "\telse NULL end as %s\n" % columnname
            columns.append(case)
            out_table_info.add_column(ColumnInfo(columnname, datatype))

    if len(out_table_info.columns) != len (columns):
        raise Exception("New table has wrong number of columns: %s != %s" % (str(out_table_info), columns))

    from_clause = "%s.%s\n" % (src_schema, main_table.table)
    for table in schema_info.join_tables:
        # AL uses user_id as distkey and received_at as sortkey
        from_clause += "left outer join %s.%s using (id)\n" % (src_schema, table)

    select = "select %s from %s %s" % (",".join(columns), from_clause, where)
    return select, out_table_info

parser = argparse.ArgumentParser(description='Copies segment data from a segment redshift warehouse to BG redshift')
parser.add_argument('--src_url', type=str, help='segment db url', required=True)
parser.add_argument('--dst_url', type=str, help = 'Bytegain redshift DB URL', required = True)
parser.add_argument('--schema_desc', type = str, help = 'Filename of schema description', required = True)
parser.add_argument('--dst_schema', type = str, help = 'BG customer name', required = True)
parser.add_argument('--src_region', type = str, help = 'SRC AWS region', default = "us-east-1")
parser.add_argument('--start_date', type = segment_utils.valid_date, help = 'Start date of data to copy', default = None)
parser.add_argument('--end_date', type = segment_utils.valid_date, help = 'End date of data to copy', default = None)
parser.add_argument('--s3_key_id', type=str, default = "AKIAIGHWWWWWV534ALTA",
                    help='AWS user id for S3', required=True)
parser.add_argument('--s3_secret_key', type=str, default = os.environ.get("S3_SECRET_KEY"),
                    help='AWS secret key for S3', required=True)
parser.add_argument('--sortcolumn', type=str, default = "user_id", help = "Sort/dist column for created table")
parser.add_argument('--skip_existing', type=bool, default = False, help = "If true, then skip tables that already exist in DST")
parser.add_argument('--incremental', action="store_true", default = False, help = "If set, then incrementally loads data into tables")
parser.add_argument('--output_table', type=str, default=None, help = "Name of output table to create/add to")
parser.add_argument('--where', type=str, help = 'additional where clause', default = None)
parser.add_argument('--s3_only', action="store_true", default = False, help = 'If set, then data is only copied to S3')
parser.add_argument('--gzip', action="store_true", default = False, help = 'If set, then s3 data is GZIP')
parser.add_argument('--s3_prefix', type=str, default = None, help = "S3 Destination directory for data load")

args = parser.parse_args()

# Create Where clause from start/end dates
where = db_utils.create_date_clause(args.start_date, args.end_date)
schema_info = read_schema_from_file(args.schema_desc)

# Clone the main tracks table
src_conn_factory = db_utils.ConnectionFactory(args.src_url)
dst_conn_factory = db_utils.ConnectionFactory(args.dst_url)
events = [g[0] for g in schema_info.main_table.get_column('event')._groups]
if len(events) > 0:
    tracks_where = db_utils.and_where(where, "event in ('%s')" % ("' ,'".join(events)))
else:
    tracks_where = where
main_cloner = TableCloner(schema_info.main_table, tracks_where, args.dst_schema, args.src_region, args.s3_key_id, args.s3_secret_key, args.s3_prefix,
	use_tablename_column = True)
if args.s3_only:
    manifest = main_cloner.run_select_from_redshift(src_conn_factory.get_connection(), args.gzip)
else:
    main_cloner.clone_table(src_conn_factory, dst_conn_factory, main_table, "id", args.skip_existing, args.incremental)

page_cloner = TableCloner(schema_info.page_table, where, args.dst_schema, args.src_region, args.s3_key_id, args.s3_secret_key, args.s3_prefix,
	use_tablename_column = True)
if args.s3_only:
    manifest = page_cloner.run_select_from_redshift(src_conn_factory.get_connection(), args.gzip)
else:
    page_cloner.clone_table(src_conn_factory, dst_conn_factory, main_table, "id", args.skip_existing, args.incremental)

# Iterate through all tables in schema and clone them
for table in list(schema_info.join_tables.values()):
    cloner = TableCloner(table, where, args.dst_schema, args.src_region, args.s3_key_id, args.s3_secret_key, args.s3_prefix,
    	use_tablename_column = True)
    tablename =  "%s_%s" % (table.schema, table.table)
    if args.s3_only:
        manifest = cloner.run_select_from_redshift(src_conn_factory.get_connection(), args.gzip)
    else:
        success = cloner.clone_table(src_conn_factory, dst_conn_factory, tablename, "id", args.skip_existing, args.incremental)

save_schema("%s_s3" % (args.schema_desc), schema_info, compact = True)

# Join tracks & other tables into one big table
if args.output_table and not args.s3_only:
    dst_cur = dst_conn_factory.get_connection().cursor()
    dst_cur.execute("set search_path to '%s'" % args.dst_schema)
    schema_info.translate_schema_tablenames(args.dst_schema)
    where = db_utils.create_date_clause(args.start_date, args.end_date, schema_info.main_table.full_name)

    # TODO(chris): Filter down to just joined events.
    if args.where is not None:
        if len(where) >0:
            where = "%s AND %s" % (where, args.where)
        else:
            where = args.where

    if len(where) > 0:
        where = "WHERE %s" % where

    select, out_table_info = get_join_select(schema_info, args.dst_schema, args.output_table, where)
    save_schema("%s" % args.output_table, out_table_info, compact = True)
    create_joined_table(dst_cur, out_table_info, select, args.sortcolumn)



import argparse
import getpass
import json
import string

import requests


def create_post(client, log_files):
    attachment = {
        "fallback": "Pipeline run",
        "pretext": "Pipeline run",
        "mrkdwn_in": ["fields"],
    }
    fields = []
    for log_file in log_files:
        label, filename = log_file.split(":", 1)
        with open(filename, "r") as f:
            part = 1
            text = ""
            for line in f:
                line = string.replace(line, "    ", "\t")
                # Slack only allows 2K / attachment field, so break up long files
                if len(text) + len(line) > 1990:
                    fields.append({
                        "title": "%s [part %d]" % (label, part),
                        "value": "```%s```" % text,
                        "short": False})
                    part += 1
                    text = ""
                text = text + line

            title = label if part == 1 else "%s [part %d]" % (label, part)
            fields.append({
                "title": title,
                "value": "```%s```" % text,
                "short": False})

    attachment["fields"] = fields
    post = {"text": "*%s* results from %s" % (client, getpass.getuser()),
            "attachments": [attachment]}
    return post


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Copies and Joins S3 data files to Bytegain S3 storage')
    parser.add_argument('--client', type=str, help='Client name', required=True)
    parser.add_argument('--slack_url', type=str, help='Slack Webhook URL',
                        default="https://hooks.slack.com/services/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")
    parser.add_argument('--log_files', type=str, help='Description:filename of logs to upload', nargs="+",
                        required=True)
    args = parser.parse_args()

    post = create_post(args.client, args.log_files)
    r = requests.post(args.slack_url, data=json.dumps(post))
    print(r.content)



from collections import defaultdict
from numbers import Number

class ColumnTracker(object):
    """A ColumnTracker tracks information about a column and the values it contains.
    The number of occurrences of the first _max_items distinct values are counted.
    """
    def __init__(self, name, max_items):
        self._name = name
        self._max_items = max_items
        self._counts = defaultdict(int)
        self._total_non_null_count = 0
        self._total_count = 0
        self._datatype = None

    def add_entry(self, value):
        try:
            if self._datatype is None and value is not None:
                py_type = type(value)
                if py_type == str or py_type == str:
                    self._datatype = "varchar"
                elif py_type == bool or py_type == int:
                    self._datatype = "int"
                elif isinstance(value, Number):
                    self._datatype = "float"
                else:
                    self._datatype = str(py_type)

            if value in self._counts or len(self._counts) < self._max_items:
                self._counts[value] += 1

            self._total_count += 1
            if value is not None:
                self._total_non_null_count += 1
        except Exception as e:
            print("Could not add %s to %s: %s" % (str(value), self._name, str(e)))
            pass

    def get_top_values(self, min_percent):
        values = []
        min_count = self._total_count * min_percent
        for key, value in sorted(iter(list(self._counts.items())), key=lambda k_v: (k_v[1],k_v[0]), reverse = True):
            if value >= min_count and self._total_count > 0:
                values.append((key, value * 100.0 / self._total_count))

        return values

class TableTracker(object):
    """A TableTracker is a collection of ColumnTrackers for each column in a table.
    """
    def __init__(self, name, max_column_items, is_tracks = False):
        self._name = name
        self._max_column_items = max_column_items
        self._columns = {}
        self._total_count = 0
        self._no_userid  = 0
        self._is_tracks = is_tracks

    def _add_column(self, column_name, value):
        if column_name not in self._columns:
            self._columns[column_name] = ColumnTracker(column_name, self._max_column_items)
        self._columns[column_name].add_entry(str(value, "UTF-8") if type(value) == str else value)

    def _add_columns(self, iteritems, prefix = None):
        for column_name, value in iteritems:
            full_name = ("%s.%s" % (prefix, column_name)) if prefix else column_name
            if type(value) == dict:
                self._add_columns(iter(list(value.items())), prefix = full_name)
            elif type(value) in [list, tuple]:
                for entry in value:
                  self._add_columns([(column_name, entry)], prefix = prefix)                    
            else:
                self._add_column(full_name, value)

    def add_entry(self, values, has_userid):
        self._total_count += 1
        if not has_userid:
            self._no_userid += 1

        if self._is_tracks:
            self._add_columns([("event", values["event"])])
            self._add_columns(iter(list(values["context"].items())))
        else:
            self._add_columns(iter(list(values.items())))

    def print_pretty(self, min_percent):
        if self._total_count > 0:
            print("%s: %d [user_id: %3f%%]" % (self._name, self._total_count, 100 - (self._no_userid * 100.0 / self._total_count)))
        for key, column in sorted(iter(list(self._columns.items())), key=lambda k_v1: (k_v1[1]._total_count, k_v1[0]), reverse = True):
            if column._datatype is not None:
                if column._total_count > self._total_count * 0.02:
                    top_values = column.get_top_values(min_percent)
                    top_values = ["%s: %d%%" % (x[0], x[1]) for x in top_values]
                    print("    %14s: %s" % (key, str(top_values)))



import pyximport

pyximport.install()

import sys
import tensorflow as tf
import bytegain.custom.cust.apartmentlist.al_xgb as al_xgb
import bytegain.custom.cust.apartmentlist.leads_filter_170 as lf

feature_file = 'bytegain/cust/apartmentlist/features/al_1.7.0.json'


def _int64_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))

def _byte_feature(value):
    return tf.train.Feature(byte_list=tf.train.BytesList(value=value))

def _float_array_feature(value):
    return tf.train.Feature(float_list=tf.train.FloatList(value=value))

def create_records(filename_base, dataset):
    writer = None
    compression_type = tf.python_io.TFRecordCompressionType.GZIP
    options=tf.python_io.TFRecordOptions(compression_type=compression_type)

    for i in range(0, len(dataset._data)):
        if i % 10000 == 0:
            sys.stdout.write("writing %d    \r" % i)
            sys.stdout.flush()
            if writer:
                writer.close()
            writer = tf.python_io.TFRecordWriter(filename_base % (i/10000), options = options)

        if len(dataset._data[i]) != rh.width:
            raise Exception("Bad width: %d" % len(dataset._data[i]))
        feature = {'label': _int64_feature(int(dataset._labels[i])),
                   'id': _int64_feature(dataset._ids[i]),
                   'data': _float_array_feature(dataset._data[i])}

        example = tf.train.Example(features=tf.train.Features(feature=feature))

        # Serialize to string and write on the file
        writer.write(example.SerializeToString())
    writer.close()


rh, datasets = al_xgb.create_rh_and_dataset(feature_file, random_seed=2828712, filter = lf.filter2,
                                            downsample_negative=False)
print("Width: %d" % rh.width)
create_records('/data3/al_tensorflow/dataset_all_train_%04d.tfrecord', datasets.train)
create_records('/data3/al_tensorflow/dataset_all_test_%04d.tfrecord', datasets.test)



"""
The Wofs class supports uploading and downloading files from write-once file systems such as
Amazon S3, Google Storage, or local files.

Filenames that begin with "s3://<bucket>/..." are on Amazon S3.
Filenames that begin with "gs://<bucket>/..." are on Google Storage.
Local files can use any name not containing "://".

A Wofs instance stores credentials use to read or write the write-once file systems via the
upload() and download() methods.
"""
import os
import shutil

import boto3
# import google.cloud.storage

# from bytegain.zap.utils import gcs
from typing import Optional

class Wofs(object):

    _KNOWN_FS = ('gs', 's3')

    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None, aws_profile_name=None,
                 gcs_project=None):
        # type: (Optional[str], Optional[str], Optional[str], Optional[str]) -> Wofs
        """
        :param aws_access_key_id: Used for access to S3
        :param aws_secret_access_key: Used for access to S3
        :param aws_profile_name: Used for access to S3
        :param gcs_project: GCS project name to use to create client
        """
        self._aws_access_key_id = aws_access_key_id
        self._aws_secret_access_key = aws_secret_access_key
        self._aws_profile_name = aws_profile_name
        self._gcs_project = gcs_project
        self._boto_session = None
        self._gcs_client = None

    def download(self, src_path, dst_local_path):
        # type: (str, str) -> None
        """
        Copies the file at src_path to dst_local_path
        :param src_path: A file name to be copied to dst_local_path
        :param dst_local_path: A local file name to be written
        :return:
        """
        fs, bucket, key = Wofs._get_fs_bucket_key(src_path)
        if fs == 'gs':
            gcs.download_to_local_file(src_path, local_file=dst_local_path, client=self._get_gcs_client())
        elif fs == 's3':
            session = self._get_boto_session()
            s3 = session.client('s3')
            with open(dst_local_path, "wb") as f:
                response = s3.get_object(Bucket=bucket, Key=key)
                body = response['Body']
                data = body.read()
                f.write(data)
        elif fs == 'local':
            shutil.copyfile(src_path, dst_local_path)
        else:
            raise Exception("Invalid fs '%s' returned by _get_fs_bucket_key()" % fs)

    def upload(self, src_local_path, dst_path):
        # type: (str, str) -> None
        """
        Copies the file at src_local_path to dst_path.
        :param src_local_path: A valid local file name to be copied to dst_path
        :param dst_path: A file name to be written
        :return:
        """
        fs, bucket, key = Wofs._get_fs_bucket_key(dst_path)
        if fs == 'gs':
            if os.path.basename(dst_path) != os.path.basename(src_local_path):
                raise RuntimeError('Wofs.upload() to a gs:// target requires identical basename of src and dst')
            gcs.upload_local_files(src_local_path, os.path.dirname(dst_path), client=self._get_gcs_client())
        elif fs == 's3':
            session = self._get_boto_session()
            s3 = session.client('s3')
            with open(src_local_path, "rb") as f:
                s3.put_object(Bucket=bucket, Key=key, Body=f)
        elif fs == 'local':
            dirname = os.path.dirname(key)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            shutil.copy(src_local_path, key )
        else:
            raise Exception("Invalid fs '%s' returned by _get_fs_bucket_key" % fs)

    def _get_gcs_client(self):
        if not self._gcs_client:
            self._gcs_client = google.cloud.storage.Client() \
                if self._gcs_project is None else google.cloud.storage.Client(project=self._gcs_project)
        return self._gcs_client

    def _get_boto_session(self):
        # type: () -> boto3.Session
        if not self._boto_session:
            self._boto_session = boto3.Session(profile_name=self._aws_profile_name,
                                               aws_access_key_id=self._aws_access_key_id,
                                               aws_secret_access_key=self._aws_secret_access_key)
        return self._boto_session

    @staticmethod
    def _get_fs_bucket_key(path):
        # type: (str) -> (str, Union[str, None], str)
        """
        Decomposes path into the filesystem and bucket (if any)
        :param path: A file path name stored as a string.
        :return: file system, bucket, key
        """
        fs_rest = path.split('://', 1)
        if len(fs_rest) == 1:
            return 'local', None, path
        fs, rest = fs_rest
        if fs not in Wofs._KNOWN_FS:
            raise Exception('Unsupported file system: %s in path %s' % (fs, path))
        bucket_key = rest.split('/', 1)
        if len(bucket_key) == 1:
            raise Exception('Missing bucket in path %s' % path)
        bucket, key = bucket_key
        return fs, bucket, key


class GlobalWofs(object):
    _default_upload_wofs = Wofs()
    _default_download_wofs = Wofs()

    @staticmethod
    def upload(src_local_path, dst_path):
        # type: (str, str) -> None
        GlobalWofs._default_upload_wofs.upload(src_local_path, dst_path)

    @staticmethod
    def default_download(src_path, dst_local_path):
        # type: (str, str) -> None
        GlobalWofs._default_download_wofs.download(src_path, dst_local_path)



import argparse
import datetime
import glob
import math
import os
import sys
import ujson as json
from collections import Counter

import flatten_json
import json_lines
from google.cloud import bigquery

from bytegain.custom.model_data import feature_analysis
from bytegain.custom.model_data.feature_analysis import Datatype, FeatureInfo, MIN_CATEGORY_FRACTION, MIN_CATEGORIES_FRACTION

# Features type enum
TYPE_LEASE = 'lease'
TYPE_SEARCH_RANK = 'search_rank'

LOG_FREQ = 10000
LOG_VERBOSE = sys.stdout.isatty()


def log_verbose(s):
    if LOG_VERBOSE:
        print(s)
        sys.stdout.flush()


# Copied from https://github.com/liyanage/python-modules/blob/master/running_stats.py
# Based on http://www.johndcook.com/standard_deviation.html, with some fixes to always use float.
class RunningStats:
    def __init__(self):
        self.n = 0
        self.old_m = 0.0
        self.new_m = 0.0
        self.old_s = 0.0
        self.new_s = 0.0

    def clear(self):
        self.n = 0

    def push(self, x):
        x = float(x)

        self.n += 1

        if self.n == 1:
            self.old_m = self.new_m = x
            self.old_s = 0.0
        else:
            self.new_m = self.old_m + (x - self.old_m) / self.n
            self.new_s = self.old_s + (x - self.old_m) * (x - self.new_m)

            self.old_m = self.new_m
            self.old_s = self.new_s

    def mean(self):
        return self.new_m if self.n else 0.0

    def variance(self):
        return self.new_s / self.n if self.n > 0 else 0.0

    def standard_deviation(self):
        return math.sqrt(self.variance())


class FeatureStats(object):
    """
    A FeatureStats is used to compute statistics on a feature.  It has a feature name, a count of the number of
    values added for the feature, an inferred value type, and an average value for numeric types.
    """

    def __init__(self, name):
        """
        :param name: the name of the feature
        """
        self._name = name
        self._count = 0
        self._db_type = None
        self._value_freq = Counter()  # XXX space is O(number of distinct values)
        self._running_stats = RunningStats()  # to compute stddev for number features
        self._median = None

    def add_value(self, value):
        """
        :param value: the value of the feature
        :return:
        """
        if type(value) == float:
            value = round(value, 4)

        self._count += 1

        if self._db_type is None and value is not None:
            if isinstance(value, str):
                self._db_type = str
            elif isinstance(value, bool):
                self._db_type = bool
            elif isinstance(value, int):
                self._db_type = int
            elif isinstance(value, float):
                self._db_type = float
            elif isinstance(value, datetime.datetime):
                self._db_type = datetime.datetime
            else:
                print('Unknown type: %s for %s' % (str(value), self._name))
                self._db_type = 0
        if self._db_type in (int, float) and isinstance(value, (int, float)):
            self._running_stats.push(value)
        self._value_freq[value] += 1

    def name(self):
        return self._name

    def total_count(self):
        return self._count

    def adjust_count(self, row_count):
        # _value_freq has a count for None only for explicit null values that were sent in a request, but missing
        # values are assumed null too so they should be counted up to row_count
        null_count = row_count - (self._count - self._value_freq[None])
        if null_count > 0:
            self._value_freq[None] = null_count
        # update total count to be total rows
        self._count = row_count

    def null_count(self):
        return self._value_freq[None]

    def not_null_count(self):
        return self.total_count() - self.null_count()

    def db_type(self):
        return self._db_type

    def average(self):
        return self._running_stats.mean()

    def median(self):
        if self._median is not None:
            return self._median
        if self._db_type in (int, float) and self.total_count() > 0:
            vals = list(self._value_freq.items())
            vals.sort() # sorts by key, but None may be first so remove it
            if vals[0][0] is None:
                vals = vals[1:]
            half_count = self.not_null_count() / 2
            sum = 0
            for val in vals:
                sum += val[1]
                if sum >= half_count:
                    self._median = val[0]
                    break
        return self._median

    def standard_deviation(self):
        return self._running_stats.standard_deviation()

    def sorted_value_frequencies(self):
        # return a list of tuples (key, value) by value desc.
        return self._value_freq.most_common()


def _process_feature(stats, row_count):
    """
    :param stats: a FeatureStats for a feature stored in db for user/model specified by data_info
    :param row_count: (int) the total number of entries in db for any feature matching data_info
    :return: a FeatureInfo
    """
    min_count = max(row_count * MIN_CATEGORY_FRACTION, 20)
    values = []
    category_sum = 0.0
    for value, count in stats.sorted_value_frequencies():
        if count > min_count:
            percent = count / float(row_count)
            values.append([value, percent, 0.0])
            category_sum += percent
        else:
            break

    if stats.db_type() in (bool, str):
        datatype = Datatype(stats.db_type(), len(values) > 1)
    elif stats.db_type() == float:
        # For now always assume int is non-categorical
        # TODO(jjs): Add normality test for ints to distinguish between categorical vs value
        if category_sum > 0.5 and len(values) > 1:
            is_category = True
        elif category_sum > MIN_CATEGORIES_FRACTION and len(values) > 2:
            is_category = True
        else:
            is_category = False
        datatype = Datatype(stats.db_type(), is_category)
    else:
        datatype = Datatype(stats.db_type(), False)

    null_fraction = stats.null_count() / float(row_count) if row_count > 0 else 0.0
    return FeatureInfo(stats.name(), datatype, values, 0.0, stats.average(), null_fraction,
                       stats.standard_deviation(), stats.median())


class StatsRunner:
    def __init__(self, row_generator):
        self._row_generator = row_generator

    def run(self):
        feature_stats = {}  # maps feature name to FeatureStats
        row_count = 0
        for row in self._row_generator:
            row_count += 1
            features = flatten_json.flatten(row)
            for name, value in list(features.items()):
                stats = feature_stats.get(name)
                if not stats:
                    stats = FeatureStats(name)
                    feature_stats[name] = stats
                stats.add_value(value)

        print('>>> Rows: %d' % row_count)

        features = []
        for name, stats in list(feature_stats.items()):
            stats.adjust_count(row_count)
            features.append(_process_feature(stats, row_count))

        features.sort(key=lambda row: row._column)
        features.sort(key=lambda row: row._datatype._is_cat)
        features.sort(key=lambda row: row._datatype._datatype)
        features.sort(key=lambda row: row._correlation, reverse=True)
        features.sort(key=lambda row: row.is_suitable(), reverse=True)
        if LOG_VERBOSE:
            is_suitable = True
            for feature in features:
                if is_suitable and not feature.is_suitable():
                    print ("\n--------------- EXCLUDED --------------\n")
                    is_suitable = False
                print("%s" % str(feature))

        return features


# Row generator from jsonl.gz files in local_dir.
def rows_from_jsonl_files(features_type, local_dir):
    row_count = 0
    if features_type == TYPE_LEASE:
        for infile in glob.glob(local_dir + '/*.jsonl.gz'):
            log_verbose('>>> Processing file: %s' % infile)
            with json_lines.open(infile) as f:
                for item in f:
                    row_count += 1
                    if row_count % LOG_FREQ == 0:
                        log_verbose('Rows: %d' % row_count)
                    row = json.loads(item['features'])
                    yield row
    elif features_type == TYPE_SEARCH_RANK:
        request_count = 0
        service_count = 0
        for infile in glob.glob(local_dir + '/*.jsonl.gz'):
            log_verbose('>>> Processing file: %s' % infile)
            with json_lines.open(infile) as f:
                for item in f:
                    request_count += 1
                    req_json = item['request']
                    services = json.loads(req_json)['services']
                    for service in services:
                        service_count += 1
                        user_data = service['user']
                        units_data = service['units']
                        for unit in units_data:
                            row_count += 1
                            row = unit.copy()
                            row.update(user_data)
                            yield row
                    if request_count % LOG_FREQ == 0:
                        log_verbose('Requests: %d, Services: %d, Rows: %d' % (request_count, service_count, row_count))

        print('>>> Requests: %d, Services: %d, Rows: %d' % (request_count, service_count, row_count))
    else:
        raise Exception('Unknown features_type: ' + features_type)


# Row generator from BQ query.
def rows_from_bq_query(features_type, query_where):
    if features_type == TYPE_LEASE:
        query = """SELECT features FROM `bg-rest-api.al.service_logs` WHERE """ + query_where
    elif features_type == TYPE_SEARCH_RANK:
        query = """SELECT request FROM `bg-rest-api.al.search_rank_logs` WHERE """ + query_where
    else:
        raise Exception('Unknown features_type: ' + features_type)

    print('>>> Query: %s' % query)

    bq_client = bigquery.Client(project='bg-rest-api')
    job_config = bigquery.QueryJobConfig()
    job_config.use_query_cache = True
    rows_iterator = bq_client.query_rows(query, job_config=job_config)

    row_count = 0
    if features_type == TYPE_LEASE:
        for qr in rows_iterator:
            features_json = qr[0]
            row = json.loads(features_json)
            row_count += 1
            if row_count % LOG_FREQ == 0:
                log_verbose('Rows: %d' % row_count)
            yield row
    elif features_type == TYPE_SEARCH_RANK:
        request_count = 0
        service_count = 0
        for qr in rows_iterator:
            request_count += 1
            req_json = qr[0]
            services = json.loads(req_json)['services']
            for service in services:
                service_count += 1
                user_data = service['user']
                units_data = service['units']
                for unit in units_data:
                    row_count += 1
                    row = unit.copy()
                    row.update(user_data)
                    yield row
            if request_count % LOG_FREQ == 0:
                log_verbose('Requests: %d, Services: %d, Rows: %d' % (request_count, service_count, row_count))

        print('>>> Requests: %d, Services: %d, Rows: %d' % (request_count, service_count, row_count))
    else:
        raise Exception('Unknown features_type: ' + features_type)


if __name__ == '__main__':
    LOG_VERBOSE = True
    parser = argparse.ArgumentParser(description='Analyze features from AL request logs.')
    parser.add_argument('--type', help='Type of features: lease, search_rank',
                        choices=[TYPE_LEASE, TYPE_SEARCH_RANK], required=True)
    where = parser.add_mutually_exclusive_group(required=True)
    where.add_argument('--local-dir', dest='local_dir', help="Local directory containing JSONL files")
    where.add_argument('--query-where', dest='query_where',
                       help='Standard SQL query WHERE clause to run against BigQuery')
    parser.add_argument('--output-file', dest='output_file', required=True,
                        help='Local file to save feature analysis to')
    args = parser.parse_args()

    row_generator = None
    if args.local_dir:
        local_dir = os.path.abspath(args.local_dir)
        print('Local dir: %s' % local_dir)
        row_generator = rows_from_jsonl_files(args.type, local_dir)
    else:
        # print('Query WHERE: %s' % args.query_where)
        row_generator = rows_from_bq_query(args.type, args.query_where)

    stats_runner = StatsRunner(row_generator)
    features = stats_runner.run()
    feature_analysis.save_features(args.output_file, features)

#
# WHERE clause examples:
#   _partitiontime = TIMESTAMP('2017-10-23')
# Inclusive range (2 days):
#   _partitiontime BETWEEN TIMESTAMP('2017-10-23') AND TIMESTAMP('2017-10-24')
#



import psycopg2
import argparse
import time
import sys
import boto3
import os
import gzip
from datetime import date

class TableCloner(object):
    def __init__(self, table_info, where, client, src_region, s3_id, s3_secret_key, dest_dir = None, use_tablename_column = False):
        self._table_info = table_info
        self._src_schema = table_info.schema
        self._src_full_table = "%s.%s" % (self._src_schema, self._table_info.table)
        self._where = where
        self._s3_key_id = s3_id
        self._s3_secret_key = s3_secret_key
        self._client = client
        self._now = date.today()
        self._src_region = src_region
        self._dest_dir = dest_dir
        self._use_tablename_column = use_tablename_column

        if src_region == "us-east-1":
            self._s3_bucket = "bg-incoming-east"
        else:
            # TODO(jjs): Handle other regions
            self._s3_bucket = "bg-incoming-data"

    def get_s3_connection(self):
        return boto3.client('s3',
                aws_access_key_id=self._s3_key_id,
                aws_secret_access_key=self._s3_secret_key)

    def get_s3_dir(self):
        if self._dest_dir:
            return self._dest_dir
        else:
            date = self._now.strftime("%Y%m%d")
            return "%s/%s_data/%s" % (self._client, self._src_full_table, date)

    def get_where(self):
        return self._table_info.get_full_where(("WHERE %s" % self._where) if self._where else None)

    def get_select(self):   
        columns = ['"%s"' % x.name for x in self._table_info.columns]
        if self._use_tablename_column:
            columns.append("'%s'" % self._table_info.table)
        columns = ",".join(columns)
        return "select %s from %s %s" % (columns, self._src_full_table, self.get_where())

    def run_select_from_redshift(self, src_conn, gzip):
        src_cur = src_conn.cursor()
        start_time = time.time()
        statement = self.get_select()
        statement = statement.replace("'", "\\'")
        dest_csv = "%s/%s." % (self.get_s3_dir(), self._src_full_table)
        manifest = dest_csv + "manifest"
        statement = """
        unload('%s') to 's3://%s/%s'
        credentials 'aws_access_key_id=AKIAIGHWWWWWV534ALTA;aws_secret_access_key=%s'
        ALLOWOVERWRITE PARALLEL ON %s ESCAPE MANIFEST
        """ % (statement, self._s3_bucket, dest_csv, self._s3_secret_key, "GZIP" if gzip else "")

        # print ("statement: %s" % statement)
        src_cur.execute(statement)
        sys.stdout.write(("Unloaded %s in %ds. " % (self._src_full_table, time.time() - start_time)))
        sys.stdout.flush()
        self._table_info.set_s3_manifest("s3://%s/%s" % (self._s3_bucket, manifest))
        return manifest

    def copy_from_s3(self, dst_conn, dst_table, manifest):
        manifest_url = 's3://%s/%s' % (self._s3_bucket, manifest)
        credentials = 'aws_iam_role=arn:aws:iam::087456767550:role/redshift-data-loader'
        upload_statement = "copy %s from '%s' credentials '%s' MANIFEST ESCAPE GZIP COMPUPDATE ON EMPTYASNULL REGION '%s'" % (
                dst_table, manifest_url, credentials, self._src_region)
        upload_start = time.time()
        dst_cur = dst_conn.cursor()
        dst_cur.execute(upload_statement)
        sys.stdout.write(("Uploaded %s in: %ds.  " % (dst_table, time.time() - upload_start)))
        sys.stdout.flush()

    def clone_table(self, src_conn_factory, dst_conn_factory, dst_table_base, sort_key, skip_existing, incremental = False):
        sys.stdout.write("Working on: %s.. " % self._table_info.table)
        sys.stdout.flush()

        total_time = time.time()

        # Client is schema name
        dst_table = "%s.%s" % (self._client, dst_table_base)
        if skip_existing:
            try:
                dst_cur.execute("select 1 from %s" % dst_table)
                sys.stdout.write("Already exists... Skipping\n")
                return
            except:
                pass

        src_cur = src_conn_factory.get_connection().cursor()
        src_schema = self._src_schema

        col_formats = []
        for column in self._table_info.columns:
            col_desc = '%s %s' % (column.name, column.datatype)
            col_formats.append(col_desc)

        dist_text = ("distkey(%s)" % sort_key) if sort_key else ""
        sort_text = ("sortkey(%s)" % sort_key) if sort_key else ""
        if not incremental:
            dst_cur = dst_conn_factory.get_connection().cursor()
            dst_cur.execute("drop table if exists %s" % dst_table)
            create_str = "create table %s (%s) %s %s" % (dst_table, ', '.join(col_formats), dist_text, sort_text)
            # print ("Executing: %s" % create_str)
            dst_cur.execute(create_str)

        manifest = self.run_select_from_redshift(src_conn_factory.get_connection())
        self.copy_from_s3(dst_conn_factory.get_connection(), dst_table, manifest)

        print("Total clone time: %ds" % (time.time() - total_time))
        return True



import queue
from threading import Thread
import time
import numpy

"""
Returns a batch of data from dataset.  Maybe shorter if not enough available data
"""
def get_batch(dataset, width, size):
    values = [[] for i in range(width)]
    start_time = time.time()
    try:
        for x in range(0, size):
            value = next(dataset)
            for i in range(len(value)):
                values[i].append(value[i])
    except StopIteration:
        print("Got stop: " + str(len(values[0])))
        # Only stop if there's no data
        if len(values[0]) == 0:
            raise
    end_time = time.time()
    print("Batch time (%d): %5.6f" % (size, end_time - start_time))
    return values

class BatchIterator(object):
    def __init__(self, source, size, limit = None):
        self._source = source
        self._size = size
        self._done = False
        self._limit = limit
        self._left = limit

    def __iter__(self):
        return self

    def __next__(self):
        if self._done:
            raise StopIteration

        if self._limit is not None:
            if self._left == 0:
                self._done = True
                raise StopIteration

            if self._left < self._size:
                self._done = True
                return get_batch(self._source, 3, self._left)
            else:
                self._left -= self._size

        batch = get_batch(self._source, 3, self._size)
        if len(batch[0]) != self._size:
            self._done = True
        return batch

class Source(object):
    def __init__(self, row_handler, train_function, test_function, width):
        self._train_function = train_function
        self._test_function = test_function
        self._width = width
        self._row_handler = row_handler

    def all_test_data(self, batch_size):
        return self._test_function(batch_size)

    def train_data(self, batch_size):
        return self._train_function(batch_size)

class _background_consumer(Thread):
    """Will fill the queue with content of the source in a separate thread.
    """
    def __init__(self, sentinel, queue, source, batch_size = 1000):
        Thread.__init__(self)
        self._queue = queue
        self._source = source
        self._sentinel = sentinel
        self._stopped = False
        self._batch_size = batch_size

    def _add_to_queue(self, datas, labels, ids):
        datas = numpy.array(datas, dtype = numpy.float)
        # print "Adding queue entry"
        if self._queue.qsize() == self._batch_size:
            print ("Queue Full")
        self._queue.put([datas, labels, ids])

    def run(self):
        count = 0
        datas = []
        labels = []
        ids = []
        for data, label, id in self._source:
            datas.append(data)
            labels.append(label)
            ids.append(id)
            if len(datas) == self._batch_size:
                self._add_to_queue(datas, labels, ids)
                datas = []
                labels = []
                ids = []
            if self._stopped:
                break

        if len(datas) > 0:
            self._add_to_queue(datas, labels, ids)
        # Signal the consumer we are done.
        self._queue.put(self._sentinel)

    def stop(self):
        self._stopped = True
        if not self._queue.empty():
            self._queue.get(False)

class BufferedDataset(object):
    """Buffers content of an iterator polling the contents of the given
    iterator in a separate thread.
    When the consumer is faster than many producers, this kind of
    concurrency and buffering makes sense.

    The size parameter is the number of elements to buffer.

    The source must be threadsafe.
    """
    def __init__(self, queue_size, batch_size, source):
            # Object used by _background_consumer to signal the source is exhausted
            # to the main thread.
        self._sentinel = object()
        self._queue = queue.Queue(queue_size)
        self._poller = _background_consumer(self._sentinel, self._queue, source, batch_size)
        self._poller.daemon = True
        self._poller.start()
        self._count = 0

    def __del__(self):
        self._poller.stop()

    def close(self):
        self._poller.stop()

    def __iter__(self):
        return self

    def __next__(self):
        # if self._queue.qsize() == 0:
        #       print ("Queue is empty")
        item = self._queue.get(True)
        self._count += 1

        if item is self._sentinel:
            raise StopIteration()
        return item



import bytegain.nn_tools.auc_plot as auc_plot
import time
from threading import Thread
import abc
import numpy as np
import xgboost as xgb
import bytegain.custom.nn_tools.xgb_stats as xgb_stats


class BaseBoostClassifier(object):
    def __init__(self, loaded_classifier=None, group_function=None):
        self._group_function = group_function
        self._classifier = loaded_classifier

    @abc.abstractmethod
    def train(self, datasets, **args):
        pass

    @abc.abstractmethod
    def predict(self, dataset, **args):
        pass

    def get_sorted_results(self, dataset):
        preds = self.predict(dataset)
        results = []
        if self._group_function is None:
            for entry in zip(preds, dataset._labels, dataset._ids):
                results.append({"prob": entry[0], "pos": entry[1], "id": entry[2]})
            auc_plot.sort_results(results)
        else:
            index = 0
            for group_size in self._group_function(dataset):
                end = index + group_size
                group = []
                for i in range(index, end):
                    group.append({"prob": preds[i], "pos": dataset._labels[i], "id": dataset._ids[i]})
                index = end
                auc_plot.sort_results(group)
                results.append(group)
        return results

    @abc.abstractmethod
    def feature_scores(self, row_handler):
        pass

    @abc.abstractmethod
    def save(self, filename):
        pass


class XgbClassifier(BaseBoostClassifier):
    def _create_dmatrix(self, dataset, group_function=None):
        x = dataset._data
        y = dataset._labels
        weights = dataset._weights
        matrix = xgb.DMatrix(x, label=y, weight=weights)
        if group_function is not None:
            matrix.set_group(group_function(dataset))
        return matrix

    def train(self, datasets, max_rounds=401, max_depth=2, l2_norm=1000, min_increase_rounds = 40, gamma = 1, train_step_size = 0.40,
        subsample = 0.8, min_child_weight = 1, max_delta_step = 0, **args):
        dtrain = self._create_dmatrix(datasets.train, self._group_function)
        dtest = self._create_dmatrix(datasets.test, self._group_function)
        base_rate = np.sum(datasets.train.labels) / len(datasets.train.labels)
        print('train base rate: %5.3f%%' % (base_rate * 100))
        evallist = [(dtest, 'eval'), (dtrain, 'train')]
        self._param = {'silent': True, 'objective': 'binary:logistic'}
        self._param['eval_metric'] = 'auc'
        self._param["base_score"] = base_rate
        self._param["early_stopping_rounds"] = min_increase_rounds
        self._param["gamma"] = gamma
        self._param["subsample"] = subsample
        self._param['lambda'] = l2_norm
        self._param['eta'] = train_step_size
        self._param['max_depth'] = max_depth
        self._param['min_child_weight'] = min_child_weight
        self._param['max_delta_step'] = max_delta_step
        # xgb.train returns a Booster, but let's call it 'classifier' for consistency
        self._classifier = xgb.train(self._param, dtrain, max_rounds, evallist, verbose_eval=100)

    def predict(self, dataset):
        dtest = self._create_dmatrix(dataset)
        return self._classifier.predict(dtest)

    def save(self, filename):
        self._classifier.save_model(filename)

    def dump_model(self, row_handler, filename):
        fmap = filename + ".fmap"
        self._create_feature_map(row_handler, fmap)
        self._classifier.dump_model(filename, fmap, True)

    def _create_feature_map(self, row_handler, filename):
        with open(filename, "w") as f:
            for i, bucket in enumerate(row_handler._buckets):
                f.write("%d %s q\n " % (i, bucket._name))

    def feature_scores(self, row_handler):
        return xgb_stats.compute_feature_importance(self, row_handler)


class CvRunner(Thread):
    def __init__(self, fold, dataset, net, training_cycles, train_args):
        super(CvRunner, self).__init__()
        self._net = net
        self._fold = fold
        self._dataset = dataset
        self._training_cycles = training_cycles
        self._train_args = train_args

    def run(self):
        print("Running CV fold: %d" % self._fold)
        self._net.train(datasets=self._dataset, max_rounds = self._training_cycles, **self._train_args)
        self._sorted_test = self._net.get_sorted_results(self._dataset.test)

class ClonedDatasets(object):
    def __init__(self, datasets):
        self.train = datasets.train.copy()
        self.test = datasets.test.copy()

def cv(dataset, create_net, training_cycles = 361, runs = 5, parallel = False, **train_args):
    start_time = time.time()
    auc_sum = 0
    aucs = []
    test_results = []
    runners = []
    for run in range(runs):
        clone_dataset = ClonedDatasets(dataset) if parallel else dataset
        net = create_net(threads = 3) if parallel else create_net()
        runner = CvRunner(run, clone_dataset, net, training_cycles, train_args)
        runners.append(runner)
        if parallel:
            runner.start()
        else:
            runner.run()

        if run < runs - 1:
            dataset.repartition()

    if parallel:
        for thread in runners:
            thread.join()

    for runner in runners:
        sorted_test = runner._sorted_test
        aucs.append(auc_plot.compute_auc(sorted_test))
        test_results.extend(sorted_test)

    auc_plot.sort_results(test_results)
    auc = auc_plot.compute_auc(test_results)
    print("CV time: %5.3fs" % (time.time() - start_time))
    print("Total Auc: %5.3f %s" % (auc, str([("%5.3f" % a) for a in aucs])))
    print("Average Auc: %5.3f" % float(sum(aucs)/len(aucs)))
    return test_results


import pyximport
pyximport.install()

import unittest
from bytegain.custom.model_data.row_handler import RowHandler
from bytegain.custom.model_data.bucket import Bucket, ValueBucket, StringBucket, CategoricalBucket
from bytegain.custom.model_data.extractor import RowExtractor, NonNullExtractor, AutoNormedExtractor

entries = [
        {"field1" : -65, "field2" : None, "field3" : "c", "is_null" : "asda", "null_clamp" : -40, "str_cat": "cat1"},
        {"field1" : None, "field2" : -150, "field3" : "d", "is_null" : "asda", "null_clamp" : 400, "str_cat": "cat1"},
        {"field1" : 30, "field2" : 27, "field3" : "a", "is_null" : "", "null_clamp" : 0, "str_cat": "cat8"},
        {"field1" : 300, "field2" : 3000, "field3" : None, "is_null" : None, "null_clamp" : None, "str_cat": None},
]

field1_bucket = Bucket("test1", RowExtractor("field1", -27, 247), [-60, -23, 100.2, 200, 250])
field2_bucket = ValueBucket("test2", AutoNormedExtractor("field2",  -100, None, 10, 5), has_null = False)
field3_bucket = StringBucket("test3", RowExtractor("field3"), ["a", "b", "c"], has_none = False)
field4_bucket = ValueBucket("test4", NonNullExtractor("is_null"))
field5_bucket = ValueBucket("test5", AutoNormedExtractor("null_clamp", min_val = -11, max_val=121, mu = 0, sigma = 20))
field6_bucket = CategoricalBucket("test6", RowExtractor("str_cat", 0), 0, {"cat1" : 0, "cat2" : 1, "cat3" : 2})

class TestRowHandler(unittest.TestCase):
    def test_save_one_bucket(self):
        row_handler = RowHandler()
        row_handler.add_bucket(field5_bucket)
        filename = '/tmp/test_row_handler'
        row_handler.save(filename)

        loaded_handler = RowHandler(load_from=filename)
        self.compare_handlers(row_handler, loaded_handler)
        self.assertEqual(1, len(loaded_handler._buckets))


    def test_save_all_bucket(self):
        print("min max %s %s" % (str(field5_bucket._extractor._min_val), str(field5_bucket._extractor._max_val)))
        row_handler = RowHandler()
        row_handler.add_bucket(field1_bucket)
        row_handler.add_bucket(field2_bucket)
        row_handler.add_bucket(field3_bucket)
        row_handler.add_bucket(field4_bucket)
        row_handler.add_bucket(field5_bucket)
        row_handler.add_bucket(field6_bucket)
        for entry in entries:
            row_handler.get_data_from_row(entry)

        filename = '/tmp/test_row_handler'
        row_handler.save(filename)

        loaded_handler = RowHandler(load_from=filename)
        print("load: min max %s %s" % (str(loaded_handler._buckets[4]._extractor._min_val), str(loaded_handler._buckets[4]._extractor._max_val)))
        print("Loaded handler spec:")
        print(loaded_handler.print_spec());

        self.compare_handlers(row_handler, loaded_handler)

    def compare_handlers(self, expected, actual):
        for entry in entries:
            expected_data = expected.get_data_from_row(entry)
            actual_data = actual.get_data_from_row(entry)
            # actual_data[2] =923.2
            print("actual_data: " + str(actual_data))
            self.assertEqual(actual_data, expected_data)



import time
import datetime
import sys
import json
import jsonpickle
from math import sqrt
import bytegain.custom.model_data.read_from_db as read_from_db
from bytegain.custom.model_data.bucket import CategoricalBucket, ValueBucket, StringBucket
from bytegain.custom.model_data.extractor import RowExtractor, DefaultedRowExtractor


def features_to_json(features):
    return json.dumps(json.loads(jsonpickle.encode(features)), sort_keys=True, indent=4)


def features_from_json(json_str):
    return jsonpickle.decode(json_str)


def save_features(filename, features):
    with open(filename, "w") as f:
        f.write(features_to_json(features))


def read_features(filename):
    with open(filename, "r") as f:
        json_str = f.read()
    return features_from_json(json_str)


class DataInfo(object):
    def __init__(self, selector, connection_uri=None):
        self._selector = selector
        self._connection_uri = connection_uri


class Datatype(object):
    def __init__(self, datatype, is_cat):
        self._is_cat = is_cat
        self._datatype = datatype

    def is_suitable(self):
        # bools & strings must be _is_cat
        if self._is_cat:
            return True
        if self._datatype in (bool, int, float):
            return True
        return False

    def data_type(self):
        return self._datatype

    def is_cat(self):
        return self._is_cat

    def is_number(self):
        return self._datatype in (int, int, float)

    def __str__(self):
        if self._datatype == str:
            type_str = "str"
        elif self._datatype == bool:
            type_str = "bool"
        elif self._datatype in (int, int):
            type_str = "int"
        elif self._datatype == float:
            type_str = "float"
        elif self._datatype == datetime.datetime:
            type_str = "timestamp"
        elif self._datatype == datetime.date:
            type_str = "date"
        else:
            return "Other: %s" % (str(self._datatype))

        return "%s%s" % (type_str, " (cat)" if self._is_cat else "")

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self._is_cat == other._is_cat and self._datatype == other._datatype

    def __ne__(self, other):
        return not self.__eq__(other)


class FeatureInfo(object):
    def __init__(self, column, datatype, top_values, correlation, average, null_fraction, stddev, median=0):
        self._column = column
        self._datatype = datatype
        self._top_values = top_values
        self._correlation = correlation
        self._average = average
        self._null_fraction = null_fraction
        self._stddev = stddev
        self._median = median

    def is_suitable(self):
        return self._datatype.is_suitable()

    def name(self):
        return self._column

    def data_type(self):
        return self._datatype

    def average(self):
        return self._average

    def median(self):
        return self._median

    def stddev(self):
        return self._stddev

    def top_values(self):
        return self._top_values

    def null_fraction(self):
        return self._null_fraction

    def create_bucket(self, is_category=None, one_hot=False, none_value=None):
        if not self.is_suitable():
            print("Warning- including %s which is marked as not suitable" % self._column)
        extractor = RowExtractor(self._column)
        if one_hot:
            return StringBucket(self._column, extractor, [x[0] for x in self._top_values])
        elif is_category != False and (is_category == True or self._datatype._is_cat):
            return CategoricalBucket(self._column, extractor, min_fraction=0.0001, none_value=none_value)
        else:
            # print "Using ValueBucket for %s (%f)" % (self._column, extractor)
            # extractor = AutoNormedExtractor(self._column, mu = self._average, sigma = self._stddev)
            return ValueBucket(self._column, extractor, has_null=False, default_value=none_value)

    def create_predict_script_bucket(self, is_category=None, one_hot=False, none_value=None):
        if not self.is_suitable():
            print("Warning- including %s which is marked as not suitable" % self._column)
        extractor = DefaultedRowExtractor(self._column)
        if one_hot:
            return StringBucket(self._column, extractor, [x[0] for x in self._top_values])
        elif is_category != False and (is_category == True or self._datatype._is_cat):
            bucket = CategoricalBucket(self._column, extractor, min_fraction=0.015, none_value=none_value)
            bucket.set_categories([{'name': x[0], 'count': 100 * x[1]} for x in self._top_values])
            return bucket
        else:
            # print "Using ValueBucket for %s (%f)" % (self._column, extractor)
            # extractor = AutoNormedExtractor(self._column, mu = self._average, sigma = self._stddev)
            return ValueBucket(self._column, extractor, has_null=False, default_value=none_value)

    def __str__(self):
        if self._datatype.is_cat():
            value_str = ("vals(%d):" % len(self._top_values)) + str(
                ["%s: %3.0f%%" % (str(x[0]), x[1] * 100) for x in self._top_values])
        elif self._datatype.is_number():
            value_str = "avg(%5.3f) median(%5.3f) stddev(%5.3f)" % (self.average(), self.median(), self.stddev())
        else:
            value_str = ("vals(%d)" % len(self._top_values)) +\
                        (":%s: %3.0f%%" % (str(self._top_values[0][0]), self._top_values[0][1] * 100) if self._top_values else '')
        return "%30.30s: null(%3.0f%%) %10s %5.3f %s" % (
            self._column, self._null_fraction * 100, str(self._datatype), sqrt(self._correlation) * 100, value_str)


MIN_CATEGORY_FRACTION = 0.03
MIN_CATEGORIES_FRACTION = 0.20

MIN_CATEGORY_STR_FRACTION = 0.0001
MIN_CATEGORIES_STR_FRACTION = 0.05

def _process_column(cursor, data_info, column, row_count, label_average):
    """
    Returns: a FeatureInfo
    """
    sys.stdout.write("Processing: %s                             \r" % column)
    sys.stdout.flush()
    min_count = max(row_count * MIN_CATEGORY_FRACTION, 500)
    min_str_count = max(row_count * MIN_CATEGORY_STR_FRACTION, 500)
    statement = "select \"%s\", count(*), avg(%s::int::float) from %s %s group by 1 order by 2 desc" % (
        column, data_info._selector._label_field, data_info._selector._table, data_info._selector.get_where())
    cursor.execute(statement)
    values = []
    db_type = None
    category_sum = 0
    for row in cursor:
        value = row[0]
        count = row[1]
        rate = row[2]

        if db_type == None and value is not None:
            if isinstance(value, str):
                db_type = str
            elif isinstance(value, bool):
                db_type = bool
            elif isinstance(value, int) or isinstance(value, int):
                db_type = int
            elif isinstance(value, float):
                db_type = float
            elif isinstance(value, datetime.datetime):
                db_type = datetime.datetime
                break
            elif isinstance(value, datetime.date):
                db_type = datetime.date
                break
            else:
                print("Unknown type: %s for %s" % (str(value), column))
                db_type = 0
                break

        if (db_type == str and count > min_str_count) or (count > min_count):
            percent = count / float(row_count)
            values.append([value, percent, rate])
            category_sum += percent
        elif db_type is not None:
            break

    if db_type in (bool, int, float):
        if db_type == bool:
            conversion = "::int::float"
        elif db_type == int:
            conversion = "::int"
        else:
            conversion = ""
        statement = """select avg("%s"%s),
        stddev("%s"::int::float),
        avg(case when "%s" is null then 1.0 else 0.0 end)
        from %s %s """ % (column, conversion, column, column, data_info._selector._table, data_info._selector.get_where())
    else:
        statement = "select 0, 1.0, avg(case when \"%s\" is null then 1.0 else 0.0 end) from %s %s" % (
            column, data_info._selector._table, data_info._selector.get_where())

    cursor.execute(statement)
    row = cursor.fetchone()
    avg = row[0]
    stddev = row[1]
    null_fraction = row[2]

    if db_type in (bool, str):
        datatype = Datatype(db_type, len(values) > 1)
    elif db_type == float:
        # For now always assume int is non-categorical
        # TODO(jjs): Add normality test for ints to distinguish between categorical vs value
        if category_sum > 0.5 and len(values) > 1:
            is_category = True
        elif category_sum > MIN_CATEGORIES_FRACTION and len(values) > 2:
            is_category = True
        else:
            is_category = False
        datatype = Datatype(db_type, is_category)
    else:
        datatype = Datatype(db_type, False)

    correlation = 0.0
    for value in values:
        correlation += value[1] * (label_average - value[2]) * (label_average - value[2])

    return FeatureInfo(column, datatype, values, correlation, avg, null_fraction, stddev)


def process_features(data_info):
    """
    Returns: a list of FeatureInfo
    """
    read_from_db.dates_as_timestamp(True)

    sys.stdout.write("Collecting features from: %s " % data_info._selector._table)
    sys.stdout.flush()
    start_time = time.time()
    features = []
    meta_statement = "select * from %s %s limit 1" % (data_info._selector._table, data_info._selector.get_where())
    connection = read_from_db.create_connection()
    with connection as conn:
        cursor = conn.cursor()
        cursor.execute(meta_statement)
        columns = []
        for column in cursor.description:
            columns.append(column[0].lower())

        time1 = time.time()
        sys.stdout.write(" (%5.3fs)\n" % (time1 - start_time))
        sys.stdout.write("Counting rows: ")
        sys.stdout.flush()

        meta_statement = "select count(1), avg(%s::int::float) from %s %s " % (
            data_info._selector._label_field, data_info._selector._table, data_info._selector.get_where())
        cursor.execute(meta_statement)
        row = cursor.fetchone()
        row_count = row[0]
        label_average = row[1]
        time2 = time.time()
        sys.stdout.write("%d (%5.3fs)\n" % (row_count, time2 - time1))
        for column in columns:
            features.append(_process_column(cursor, data_info, column, row_count, label_average))

    time3 = time.time()
    print("Processing features: %5.3fs" % (time3 - time2))
    print("Total time: %5.3fs" % (time3 - start_time))

    features.sort(key=lambda row: row._column)
    features.sort(key=lambda row: row._datatype._is_cat)
    features.sort(key=lambda row: row._datatype._datatype.__name__)
    features.sort(key=lambda row: row._correlation, reverse=True)
    features.sort(key=lambda row: row.is_suitable(), reverse=True)
    is_suitable = True
    for feature in features:
        if is_suitable and not feature.is_suitable():
            print ("\n--------------- EXCLUDED --------------\n")
            is_suitable = False
        print("%s" % str(feature))

    return features



"""
Transform events into user sessions.

Input should be a top level directory, e.g. 'gs://bg-dataflow/all_features' containing sub-folders
'/web_pages/' and 'lead_scoring_lead_scoring_set/'. Each of them should contain compressed .cvs
files and a pickle file with TableInfo
"""

import argparse
import json
import logging
import apache_beam as beam
from bytegain.beam.beam_util import encode_json, DoFnWithCounters, run_pipeline, create_pipeline, get_counter
from bytegain.zap.utils import gcs

from bytegain.custom.model_data.schema_from_TableInfo import get_schema

logger = logging.getLogger('events_to_sessions')


class MakeDict(DoFnWithCounters):
    """
        Convert csv rows into a dictionary and map columns to specific zap-compatible fields
    """

    def __init__(self, table_info):
        super(MakeDict, self).__init__()
        # Map keys for consistency with zap
        self._key_map = {'user_id': 'userId',
                         'anonymous_id': 'anonymousId',
                         'received_at': 'timestamp',
                         'value_at': 'timestamp',
                         'created_at': 'timestamp',
                         'context_page_title': 'title',
                         'referrer': ['context', 'page', 'referrer'],
                         'context_page_url': ['context', 'page', 'url'],
                         'context_user_agent': ['context', 'userAgent'],
                         'context_page_path': ['context', 'page', 'path'],
                         'context_page_referrer': ['context', 'page', 'referral'],
                         'context_page_search': ['context', 'page', 'search'],
                         'notified_at': 'timestamp',
                         'leased_at_notified_property': 'label',
                         'original_timestamp': 'originalTimestamp',
                         'context_library_version': ['library', 'version'],
                         'context_library_name': ['library', 'name'],
                         # 'name': 'title'
                         # it seems like 'name' is the only useful feature in '*.screens' table,
                         # let's map it to 'title'
                         }
        self._columns = []
        for column in table_info.columns:
            col_name = column.name
            if col_name in self._key_map:
                col_name = self._key_map[col_name]
            self._columns.append(col_name)

    def nested_set(self, dic, keys, value):
        for key in keys[:-1]:
            dic = dic.setdefault(key, {})
        dic[keys[-1]] = value

    def process(self, row):
        dict = self._make_dict(row)
        if dict:
            yield dict

    def _make_dict(self, row):
        import csv
        values = next(csv.reader([row.encode('utf-8')], delimiter='|', escapechar='\\'))
        values = [str(val, 'utf-8') for val in values]
        row_dict = {}
        if len(self._columns) != len(values):
            self.get_counter('bad_row').inc()
            logging.info('Bad row: %s' % row)
            return
        for c, key in enumerate(self._columns):
            if type(key) is not list:
                key = [key]
            self.nested_set(row_dict, key, values[c])
        try:
            return row_dict
        except Exception as e:
            print("Bad line: %s" % row)
            raise e


class MergeEventsFeatures(DoFnWithCounters):
    """
        Create a user session: merge events, assign types and sort them by timestamp
    """
    def __init__(self):
        super(MergeEventsFeatures, self).__init__()

    def process(self, xxx_todo_changeme):
        (user, data) = xxx_todo_changeme
        from bytegain.beam.beam_util import get_counter

        (_web_pages, _features, _acappella_leases, _ios_screens, _android_screens, _interest, _web_compass) = data
        # Make all events as 'page' for compatibility with the zap pipeline
        for p in _web_pages:
            p['type'] = 'page'

        _events = _web_pages

        for s in _ios_screens:
            s['type'] = 'screen'

        _events += _ios_screens

        for s in _android_screens:
            s['type'] = 'screen'

        _events += _android_screens

        for c in _web_compass:
            c['type'] = 'compass'

        _events += _web_compass

        # try interest as a page
        # _events += _interest

        if len(_events) == 0:
            get_counter('sessions', 'users_without_events').inc()
            return
        else:
            get_counter('sessions', 'users_with_events').inc()

        # Mark other events as different types (they will be filtered out via rows_filter)
        for f in _features:
            get_counter('session', 'feature').inc()
            f['type'] = 'features'
        for l in _acappella_leases:
            get_counter('session', 'acappella_lease').inc()
            l['type'] = 'acappella_lease'

        # Now merge features, acappella leases into events
        _events += _features
        _events += _acappella_leases

        # Sort events by timestamp
        _events = sorted(_events, key=lambda _event: _event['timestamp'])
        yield user, _events


def make_sessions(driver_input, session_generator_config, dataflow_args, output_sessions, tmp_dir):
    """
    Called by the zap driver, from the sessions stage.
    """
    p = argparse.ArgumentParser()
    p.add_argument('--redshift_data', required=(driver_input.command == 'predict'), help='GCS URL to an existing Redshift data import')
    p.add_argument('--redshift_training_data', required=True, help='GCS URL to an existing Redshift data used in '
                                                                   'training')
    p.add_argument('--sessions_num_shards', type=int, default=-1, help='Override default num_shards')
    args, _ = p.parse_known_args(driver_input.extra_args)

    redshift_training_data = args.redshift_training_data.rstrip('/')
    logger.info('Redshift training data: %s', redshift_training_data)

    redshift_data = args.redshift_data.rstrip('/') if driver_input.command == 'predict' else redshift_training_data
    logger.info('Redshift data: %s', redshift_data)

    num_shards = args.sessions_num_shards if args.sessions_num_shards >= 0 else session_generator_config['num_shards']
    session_generator_config['num_shards'] = num_shards
    logger.info('Num shards: %s', num_shards)

    # Data start and end dates come from a json file written by the data importer tool. Read the file early, if it
    # doesn't exist it's a sign the import job didn't make it to the end.
    dates = json.loads(gcs.get_file_contents_as_string(redshift_data + '/data_start_end_dates.json'))
    data_start_date = dates['data_start_date']
    assert data_start_date
    data_end_date = dates['data_end_date']
    assert data_end_date
    # Update session_generator_config so that the driver can track the dates
    session_generator_config['data_start_date'] = data_start_date
    session_generator_config['data_end_date'] = data_end_date

    # Extract schemas from training dataset
    subdirs = {
        'web_pages': '/web_pages/',
        'features': '/ml_notified_training_set/',
        'acappella_leases': '/etl_acappella_leases/',
        'ios_screens': '/ios_screens/',
        'android_screens': '/android_screens/',
        'interest': '/api_interest/',
        'web_compass': '/web_compass/'
    }
    if driver_input.command == 'predict':
        subdirs['features'] = '/ml_notified_features/'
    schemas = get_schema(redshift_data, subdirs)

    # TODO (yuri): Create buckets somewhere here

    p = create_pipeline(dataflow_args, 'al-events-to-sessions')

    tables = ['web_pages', 'features', 'acappella_leases', 'ios_screens', 'android_screens', 'interest', 'web_compass']
    data = []
    for key in tables:
        data.append(
            (p
             | 'read %s' % key >> beam.io.ReadFromText(redshift_data + subdirs[key] + '*.gz')
             | 'make %s dict' % key >> beam.ParDo(MakeDict(table_info=schemas[key]))
             | 'map %s to userId' % key >> beam.Map(lambda x: (x['userId'], x))
             )
        )

    _ = (data | beam.CoGroupByKey()
            | 'merge events' >> beam.ParDo(MergeEventsFeatures())
            | 'ignore empty and large sessions' >> beam.Filter(lambda user__events: 0 < len(user__events[1]) < 1000)
            # 1000 is an arbitrary cutoff
            | 'encode json' >> beam.Map(encode_json, counter=get_counter("make_sessions", "nonempty_sessions"))
            | 'save' >> beam.io.WriteToText(output_sessions, file_name_suffix='.gz', num_shards=num_shards)
            )

    counters = run_pipeline(p)

    return {'counters': counters,
            'redshift_data': redshift_data,
            'redshift_training_data': redshift_training_data}



import pyximport

pyximport.install()

import bytegain.custom.cust.apartmentlist.al_xgb as al_xgb
import bytegain.nn_tools.auc_plot as auc_plot
import bytegain.custom.cust.apartmentlist.leads_filter_160 as lf160
import time
import bytegain.custom.nn_tools.xgb_classifier as xgb

# Data with prediction
feature_file = 'bytegain/cust/apartmentlist/features/al_prob1.json'
# features = tfl.run_feature_analysis(feature_file)
rh, datasets = al_xgb.create_rh_and_dataset(feature_file, downsample_negative=False, random_seed=2828712, filter = lf160.filter)

n_rounds = 500

models = {'xgb': xgb.XgbClassifier()}

for model_name in list(models.keys()):
    print("Training the %s classifier"%model_name)
    time0 = time.time()
    classifier = models[model_name]
    classifier.train(datasets, max_rounds=401, max_depth=4, l2_norm=1000, subsample=0.9, train_step_size=0.3,
                     min_child_weight=100)
    print("Took %.5f sec to train %s"%(time.time() - time0, model_name))
    features = classifier.feature_scores(rh)
    print("--------------------")
    for feature in features:
        print("%25s: %4.1f" % (feature[0], feature[1]))
    results = classifier.get_sorted_results(datasets.test)
    print("--------------------")
    print("The AUC for the %s is %.5f"%(model_name, auc_plot.compute_auc(results)))
    print("--------------------")
    auc_plot.plot_auc('/tmp/auc_'+model_name+'.png', results)



""" Implements a IP -> Country lookup.
"""

import csv
import gzip
import re
#import ipaddress

from bisect import bisect_left
from collections import defaultdict

ip_parser = re.compile('^(\d+)\.(\d+)\.(\d+)\.(\d+)$')
def atonl(s):
    """ Convert a string IP to network byte order int.

    This is roughly 2x as fast as int(ipaddress.ip_address(s.decode('utf-8')))
    """
    result = re.match(ip_parser, s)
    if not result:
        return -1
    o1 = int(result.group(1))
    o2 = int(result.group(2))
    o3 = int(result.group(3))
    o4 = int(result.group(4))
    if o1 > 255 or o2 > 255 or o3 > 255 or o4 > 255:
        return -1
    return o1 << 24 | o2 << 16 | o3 << 8 | o4

class IPGeo(object):
    BAD_IP = 'BAD_IP'
    UNKNOWN_IP_RANGE = 'UNKNOWN_IP_RANGE'

    def __init__(self, path):
        """ Expects path to the file generated by create_ip_country_csv_gz().
        """
        print("Loading IP Country map...")
        with gzip.GzipFile(path) as f:
            reader = csv.reader(f)
            header = next(reader)
            assert header == ['first', 'last', 'country']
            self._ranges = [(int(row[0]), int(row[1]), row[2]) for row in reader]
        self._lasts = [row[1] for row in self._ranges]
        print("Loaded %d ranges." % len(self._ranges))

    def lookup_country(self, ip_str):
        if not ip_str:
            return self.BAD_IP
        ip = atonl(ip_str)
        if ip == -1:
            return self.BAD_IP
        index = bisect_left(self._lasts, ip)
        if index < len(self._ranges):
            entry = self._ranges[index]
            if ip >= entry[0] and ip <= entry[1]:
                return entry[2]
        return self.UNKNOWN_IP_RANGE


def create_ip_country_csv_gz(country_locations_path, country_blocks_path, out_path):
    geoid_to_country = {}
    with open(country_locations_path) as f:
        reader = csv.reader(f)
        header = next(reader)
        assert header[0] == 'geoname_id'
        assert header[4] == 'country_iso_code'
        for row in reader:
            geoid_to_country[row[0]] = row[4]

    ranges = []
    with open(country_blocks_path) as f:
        reader = csv.reader(f)
        header = next(reader)
        assert header[0] == 'network'
        # TODO(chris): Should I be using header[3] == represented_country_geoname_id instead?
        assert header[2] == 'registered_country_geoname_id'
        for row in reader:
            if not row[2]:
                continue
            network = ipaddress.ip_network(row[0].decode('utf-8'))
            ranges.append([int(network.network_address), int(network.broadcast_address), geoid_to_country[row[2]]])

        ranges = sorted(ranges, key=lambda x: x[0])

        with gzip.GzipFile(out_path, 'w') as f:
            writer = csv.writer(f)
            writer.writerow(['first', 'last', 'country'])
            for row in ranges:
                writer.writerow(row)



from google.cloud import bigquery
from bytegain.custom.model_data.schema_from_TableInfo import get_schema


TYPE_MAP = {'timestamp without time zone': 'TIMESTAMP',
            'character varying(256)': 'STRING'}


def if_table_exists(client, table_ref):
    from google.cloud.exceptions import NotFound
    try:
        client.get_table(table_ref)
        return True
    except NotFound:
        return False


def create_table(project, dataset_id, table_id, source_uris, schema_file_uri,
                 column_map, source_format='CSV', delimiter='|'):
    """
    Creates `table_id` in `dataset_id` without `project` from files `source_uris` using table_info file
    located at `schema_file_uri`.

    :param (str) project: GCS project
    :param (str) dataset_id: Dataset in `project`
    :param (str) table_id: Table in dataset `dataset_id`
    :param (str) source_uris: GCS path to files that will be exported into `table_id`
    :param (str) schema_file_uri: TableInfo file used to extract schema for the table
    :param (str) source_format: Source format, e.g. 'CSV'
    :param (str) delimiter: Delimiter, e.g. '|'
    :param (dict[str, dict]) column_map:
    """
    client = bigquery.Client(project=project)

    dataset_ref = client.dataset(dataset_id)
    table_ref = dataset_ref.table(table_id)

    # Try to delete the table
    if if_table_exists(client, table_ref):
        client.delete_table(table_ref)
        print("Deleted table '{}'.".format(table_id))

    # Load TableInfo
    table_info = get_schema(schema_file_uri, subdirs={'all': ''}).get('all')
    assert table_info, "Table info not found"

    # Set schema from table_info
    schema = []
    for col in table_info.columns:
        col_name = col.name
        col_type = col.datatype
        # Map types
        col_type = TYPE_MAP.get(col_type, col_type)
        col_map = column_map.get(col_name)
        # Map names and types if needs to be converted
        if col_map:
            field_name = col_map.get('name', col_name)
            field_type = col_map.get('type', col_type)
        else:
            field_name = col_name
            field_type = col_type
        schema.append(bigquery.SchemaField(field_name, field_type))

    job_config = bigquery.LoadJobConfig()
    job_config.schema = schema
    job_config.field_delimiter = delimiter

    format = None
    for allowed_format in vars(bigquery.SourceFormat):
        if source_format==allowed_format:
            format = source_format
    assert format, "'%s' is not allowed format" % source_format

    job_config.source_format = format
    load_job = client.load_table_from_uri(
        source_uris, table_ref, job_config=job_config
    )  # API request
    print("Starting job {}".format(load_job.job_id))

    load_job.result()  # Waits for table load to complete.
    print("Job finished.")

    destination_table = client.get_table(table_ref)
    print("Loaded {} rows.".format(destination_table.num_rows))



import argparse
import csv
import tempfile

from bytegain.zap.utils import gcs
from google.cloud import storage


def _extract_date(filename):
    return filename.rsplit('/', 1)[-1].split('_')[4]


def combine_predictions_csv(gs_path, output, project='tough-hologram-156002'):
    """
    Reads predictions csv files from GCP and writes them into one big csv

    :param gs_path: path to csv files with predictions
    :param output: local path to a big csv
    :param project: GCP project
    :return:
    """
    print('Getting list of files from GCP')
    gcs_client = storage.Client(project=project)
    bucket, path = gcs.parse_gs_url(gs_path)
    bucket = gcs_client.bucket(bucket)
    blobs = list(bucket.list_blobs(prefix=path))
    blobs_csv = []
    # Get all blobs with predictions
    for blob in blobs:
        filename = blob.name
        if filename.endswith('with_emails.csv'):
            # Don't include predictions made in January (some data was missing)!
            month = _extract_date(filename).split('-')[1]
            if not month.startswith('01'):
                blobs_csv.append(blob)

    print('Sorting files')
    blobs_csv_sorted = sorted(blobs_csv, key=lambda x: x.name)

    local_tmp_path = tempfile.mkdtemp() + '/'
    from_date = '-'.join(_extract_date(blobs_csv_sorted[0].name).split('-')[:3])
    to_date = '-'.join(_extract_date(blobs_csv_sorted[-1].name).split('-')[-3:])
    output_filename = '%s_%s_%s.csv' % (output, from_date, to_date)
    print('Writing to %s' % output_filename)
    with open(output_filename, "w") as to_file:
        writer = csv.writer(to_file, delimiter=',')
        for blob_csv in blobs_csv_sorted:
            print('Reading file %s' % blob_csv.name)
            date = _extract_date(blob_csv.name)
            local_file_path = local_tmp_path + date + '.csv'
            blob_csv.download_to_filename(local_file_path)
            with open(local_file_path, mode='r') as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=',')
                for row in csv_reader:
                    row.append(date)
                    writer.writerow(row)


if __name__ == '__main__':
    parser = argparse.ArgumentParser('')
    parser.add_argument('--gs_dir', required=True, help='Path to gs with predictions',
                        default='gs://bg-dataflow/apartmentlist/user_value/predictions')
    parser.add_argument('--output', required=True, help='Path to local file where to save all predictions',
                        default='predictions.csv')

    args = parser.parse_args()

    combine_predictions_csv(args.gs_dir, args.output)



"""
Reads in a model dump and outputs the relative importance of each feature.
"""
import csv
import argparse
import tempfile
from collections import defaultdict

def compute_features_from_dump(model):
    with open(model, "r") as input:
        reader = csv.reader(input)
        total = 0
        feature_dict = defaultdict(float)
        for line in reader:
            if len(line) < 4:
                continue
            first = line[0]
            begin = first.index('[')
            last = first.index('<')
            feature = first[begin + 1: last]
            gain_text = line[3]
            gain = float(gain_text[5:])
            feature_dict[feature] += gain
            total += gain
    results = sorted(iter(list(feature_dict.items())), key = lambda x: x[1], reverse = True)
    results = [(x[0], x[1] * 100.0 / total) for x in results]
    return results

def compute_feature_importance(xgb_model, row_handler):
    with tempfile.NamedTemporaryFile(delete = False, suffix=".dump") as temp:
        xgb_model.dump_model(row_handler, temp.name)
        results = compute_features_from_dump(temp.name)
        return results

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--model', required=True)

    args = parser.parse_args()
    for k,v in compute_features_from_dump(args.model):
        print("%30s: %6.3f%%" % (k, v))



import argparse
import importlib

from bytegain.custom.cust.apartmentlist import al_xgb
from bytegain.custom.nn_tools import model_utils


parser = argparse.ArgumentParser(description='Trains an AL model',
                                 formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--model_config', required=True, help='filename of model_config')
parser.add_argument('--feature_analysis', action='store_true', help='If set then a new feature analysis is done')
parser.add_argument('--production', action='store_true', help='If set then push model to production')

args = parser.parse_args()

model_config = args.model_config

module = importlib.import_module('bytegain.custom.cust.apartmentlist.%s' % model_config)
config = getattr(module, 'prod_config' if args.production else 'config')

if args.feature_analysis:
    al_xgb.run_feature_analysis(config)

if args.production:
    al_xgb.train_and_push(config)
else:
    rh, dataset = al_xgb.create_rh_and_dataset(config)
    trained_model = al_xgb.train_from_dataset(config, rh, dataset)
    results_file = '%s.test_predictions.csv' % model_config
    model_utils.save_results_as_csv(results_file, 'id', 'pred',
        trained_model.get_sorted_results(dataset.test))



from bytegain.zap.prediction_instance import PredictionInstance
from bytegain.custom.cust.apartmentlist.user_value_model.instance_generator import RentInstanceGenerator

DAY_SECONDS = 24 * 3600


class LeaseInstanceGenerator(RentInstanceGenerator):
    def get_instances(self, rows):
        """
        Predict if a user will sign a lease at a given property.

        Label is extracted from features as follows: label = extra_features.get('leased_at_property') == 't'
        One training example per set of features.
        Use up to 7 days of data after the label.

        :param rows: list[dict], a user session
        :return: instances: list[PredictionInstance]
        """
        if len(rows) == 0:
            return []

        from bytegain.beam.beam_util import get_counter
        from bytegain.zap.abstract_instance_generator import timestamp_to_epoch, get_timestamp

        # Find all sets of features
        all_extra_features = [row for row in rows if row.get('type') == 'features']
        get_counter('instance', 'features_found').inc(len(all_extra_features))

        get_counter('instance', 'users').inc()

        instances = []

        # Make sure the number of feature buckets is the same
        default_buckets = [0] * self._bucketer.total_length

        # Predict mode - use the most recent features and construct one instance without a length limit
        if self._predict_mode:
            extra_features_buckets = default_buckets
            label = False
            if all_extra_features:
                extra_features = all_extra_features[-1]
                extra_features_buckets = self._bucketer.sample_to_vec(extra_features)
                label = extra_features.get('leased_at_property') == 't'
            # We don't need to look for the last page event, rows_filter will do it later
            if not label:
                instances.append(PredictionInstance(0, len(rows), label, predict_index=None, final_dwell=None,
                                                    future_data=None, extra_features=extra_features_buckets))
                get_counter('instance_generator', 'instance_generated').inc()
        else:
            latest_page_i = None
            # In training mode make instance per set of features
            if all_extra_features:
                features_i = 0
                timestamp_features = timestamp_to_epoch(get_timestamp(all_extra_features[features_i]))
                for row_i, row in enumerate(rows):
                    if row.get('type') == 'page':
                        latest_page_i = row_i
                    timestamp_row = timestamp_to_epoch(get_timestamp(rows[row_i]))
                    # Generate instance using up to 1 week of data after the label
                    if (timestamp_row - timestamp_features > 7 * DAY_SECONDS or row_i == len(rows) - 1) and latest_page_i:
                        label = all_extra_features[features_i].get('leased_at_property') == 't'
                        extra_features_buckets = self._bucketer.sample_to_vec(all_extra_features[features_i])
                        instances.append(PredictionInstance(0, latest_page_i + 1, label, predict_index=None, final_dwell=None,
                                                            future_data=None, extra_features=extra_features_buckets))
                        # Now use the next set of features
                        features_i += 1
                        if features_i == len(all_extra_features):
                            break
                        else:
                            timestamp_features = timestamp_to_epoch(get_timestamp(all_extra_features[features_i]))
        return instances


class LeaseInstanceGeneratorAfterFeatures(RentInstanceGenerator):
    def get_instances(self, rows):
        """
        NOTE: PREDICT MODE IS NOT IMPLEMENTED!

        Predict if a user will sign a lease at a given property.

        Label is extracted from features as follows: label = extra_features.get('leased_at_property') == 't'
        One training example per set of features.
        Use up to 7 days of data after the label.

        :param rows: list[dict], a user session
        :return: instances: list[PredictionInstance]
        """
        from bytegain.beam.beam_util import get_counter

        instances = []
        # Create instances only at the following events
        instance_events = {'page', 'screen', 'compass'}

        # Make sure the number of feature buckets is the same
        default_buckets = [0] * self._bucketer.total_length

        extra_features = {}
        for row_i, row in enumerate(rows):
            event_type = row.get('type')
            if event_type == 'compass':
                same_property = row.get('rental_id') == extra_features.get('rental_id')
                rows[row_i]['same_property'] = same_property
                get_counter('instance_generator', 'same_property_%s' % same_property).inc()
            if event_type == 'features':
                extra_features = row
                label = extra_features.get('leased_at_property') == 't'
                extra_features_buckets = self._bucketer.sample_to_vec(extra_features)
                get_counter('instance', 'features_found').inc()
                instance_begin = row_i
                instance_end = self.find_instance_end(rows=rows, rows_start=row_i,
                                                      days=7, instance_events=instance_events)
                if instance_end:
                    instances.append(PredictionInstance(instance_begin, instance_end + 1, label,
                                                        predict_index=None, final_dwell=None,
                                                        future_data=None, extra_features=extra_features_buckets))

        get_counter('instance', 'users').inc()

        # TODO (yuri): Implement predict mode if needed

        # # Predict mode - use the most recent features and construct one instance without a length limit
        # if self._predict_mode:
        #     extra_features_buckets = default_buckets
        #     label = False
        #     if all_extra_features:
        #         extra_features = all_extra_features[-1]
        #         extra_features_buckets = self._bucketer.sample_to_vec(extra_features)
        #         label = extra_features.get('leased_at_property') == 't'
        #     # We don't need to look for the last page event, rows_filter will do it later
        #     if not label:
        #         instances = [PredictionInstance(0, len(rows), label, predict_index=None, final_dwell=None,
        #                                             future_data=None, extra_features=extra_features_buckets)]
        #         get_counter('instance_generator', 'instance_generated').inc()
        return instances

    @staticmethod
    def find_instance_end(rows, rows_start, days, instance_events):
        from bytegain.zap.abstract_instance_generator import timestamp_to_epoch, get_timestamp

        latest_instance_page = None
        instance_end = None
        timestamp_features = timestamp_to_epoch(get_timestamp(rows[rows_start]))
        for row_i in range(rows_start + 1, len(rows)):
            if (timestamp_to_epoch(get_timestamp(rows[row_i])) - timestamp_features >
                        days * DAY_SECONDS or row_i == len(rows) - 1) and latest_instance_page:
                instance_end = latest_instance_page
                break
            if rows[row_i].get('type') in instance_events:
                latest_instance_page = row_i
        return instance_end



import logging
import random

from bytegain.zap.prediction_instance import PredictionInstance

DAY = 24 * 3600


class RentInstanceGenerator(object):
    def __init__(self):
        self._random = random.Random()
        self._random.seed(123456)
        self._predict_mode = False
        self._path = None
        self._bucketer = None

        self._use_days = 21
        self._blackout_days = 7
        self._signup_limit_days = 120

    def config(self, driver_input, instance_generator_config, instance_config, session_generator_data, input_sessions):
        """
        Need this function to create buckets. It is called from bytegain.zap.sessions_to_instances.make_instances
        """
        self._predict_mode = driver_input.command == 'predict'
        self._path = session_generator_data['redshift_training_data']
        assert self._path

        self._bucketer = self.get_bucketer(n_numerical_buckets=instance_config.get('n_numerical_buckets', 10))
        self._bucketer.construct_buckets()

        # Set the number of buckets generated by the bucketer. instance_config gets updated all the way
        # back up to the driver.
        instance_config['extra_features'] = self._bucketer.total_length
        instance_config['extra_features_map'] = self._bucketer.feature_map
        assert instance_config['extra_features'] > 0

    def get_bucketer(self, n_numerical_buckets):
        """
        Use schemas to construct buckets for features.

        :return:
        """
        from bytegain.custom.model_data.schema_from_TableInfo import get_schema
        from bytegain.custom.model_data.bucket_features import FeatureBucketerNonLinear as bucketer
        from bytegain.custom.cust.apartmentlist.excluded_features import EXCLUDED_FEATURES

        # Try to find pickled TableInfo file
        schemas = get_schema(self._path, {'features': '/ml_notified_training_set/'})
        # We need only schema for features
        if schemas:
            return bucketer(EXCLUDED_FEATURES, schemas['features'], n_numerical_buckets=n_numerical_buckets)
        else:
            return None
        # import pickle
        # file_features = open('bytegain/cust/apartmentlist/data/2days_median_fix/2days_median_fix/lead_scoring_lead_scoring_set_with_model/lead_scoring.lead_scoring_set_with_model..pickle')
        # schema_features = pickle.load(file_features)
        # if schema_features:
        #     return bucketer(excluded_features, schema_features)
        # else:
        #     return None

    def get_instances(self, rows):
        """
        Generate instances based on only 'extra_label': if there is no 'extra_label' - it is a negative sample.
        Make one instance per user.

        `events_to_sessions` attaches a label and features to the last event in the click stream

        :param rows:
        :return: instance
        """
        if len(rows) < 1:
            return []

        from bytegain.beam.beam_util import get_counter
        from bytegain.zap.abstract_instance_generator import timestamp_to_epoch, get_timestamp
        import random

        # Create instances only at the following events
        instance_events = {'page', 'screen', 'compass'}

        # Find the first and last page events to use
        first_page_event = None
        last_page_to_use = None
        for row_i, row in enumerate(rows):
            if row.get('type') in instance_events:
                if first_page_event is None:
                    last_page_to_use = first_page_event = row_i
                else:
                    # Save last page event before blackout
                    if timestamp_to_epoch(get_timestamp(row)) - timestamp_to_epoch(
                            get_timestamp(rows[first_page_event])) <= self._use_days * DAY:
                        last_page_to_use = row_i

        last_page_to_use_time = timestamp_to_epoch(get_timestamp(rows[last_page_to_use]))

        # Make instances only if all following criteria are met
        make_instances = True

        label_time = None
        extra_features = {}

        for row_i, row in enumerate(rows):
            if row.get('type') == 'acappella_lease':
                get_counter('instance', 'acappella_lease_found').inc()
                label_time = timestamp_to_epoch(get_timestamp(row))
                if label_time:
                    # Consider only users who sign-up no longer than 120 days since their last before blackout visit
                    sign_up_limit = self._signup_limit_days * DAY
                    if label_time - last_page_to_use_time > sign_up_limit:
                        get_counter('drop_session', 'signed_120_days_later').inc()
                        make_instances = False

                    # There should be AT LEAST 'self._blackout_days' days from the user's last event to a lease
                    blackout = self._blackout_days * DAY
                    if label_time - last_page_to_use_time < blackout:
                        get_counter('drop_session', 'events_within_blackout').inc()
                        make_instances = False

                    # Use only the earliest acappella lease
                    break

        # Use only first 'self._use_days' days of data if in a training mode
        if not self._predict_mode:
            if last_page_to_use:
                get_counter('events', 'session_truncated').inc()
                rows = rows[:last_page_to_use]

        # Make label
        label = True if label_time else False

        if self._predict_mode and label:
            make_instances = False
            get_counter('drop_user', 'user_converged').inc()

        instances = []
        # If some criteria haven't met don't construct instances
        if not make_instances:
            get_counter('instance_generator', 'no_instances').inc()
            return instances

        # Now make one instance per user

        # Make sure the number of feature buckets is the same
        default_buckets = [0] * self._bucketer.total_length

        # Find the latest features
        extra_features_buckets = default_buckets
        for row_i in range(len(rows) - 1, -1, -1):
            if rows[row_i].get('type') == 'features':
                get_counter('instance', 'features_found').inc()
                extra_features = rows[row_i]
                # Bucket features or use 'default_buckets'
                extra_features_buckets = self._bucketer.sample_to_vec(extra_features)
                break

        if len(extra_features) > 0 and random.randint(0, 100) == 73:
            logging.info("Features: %s, buckets: %s" % (extra_features, extra_features_buckets))

        instances.append(PredictionInstance(0, len(rows), label, predict_index=None, final_dwell=None,
                                            future_data=None, extra_features=extra_features_buckets))
        return instances



from bytegain.data_import.segment_utils import parse_s3_url
from bytegain.beam.beam_util import download_from_gs
import tempfile
import pickle


def read_tableinfo_from_gs(filename):
    temp = tempfile.NamedTemporaryFile()
    download_from_gs(filename, temp.name)
    return pickle.load(temp)


# Get table schema from .pickle file. One pickle file per folder is assumed
def get_schema(filespec, subdirs):
    tables_info = {}
    if filespec.startswith('gs://'):
        # Read from cloud
        from google.cloud import storage
        client = storage.Client()
        bucket_name, key, _ = parse_s3_url(filespec)
        bucket = client.get_bucket(bucket_name)
        for k, v in list(subdirs.items()):
            path = key + v
            blobs = bucket.list_blobs(prefix=path)
            for blob in blobs:
                # There should be only one pickle file per directory
                if blob.name.endswith('.pickle'):
                    pickle_string = blob.download_as_string()
                    table_info = pickle.loads(pickle_string)
                    tables_info[k] = table_info
                    break
    else:
        # Read locally
        import os
        for k, v in list(subdirs.items()):
            for file in os.listdir(filespec+v):
                if file.endswith(".pickle"):
                    tables_info[k] = pickle.load(open(filespec + v + file))
                    break
    return tables_info



import time
#from google.cloud import storage
import argparse
from bytegain.custom.data_import.table_tracker import TableTracker
from bytegain.custom.data_import.schema_info import TableInfo, ColumnInfo, SchemaInfo, save_schema

from random import Random
from collections import defaultdict
from google.cloud import bigquery
import queue
import threading

print('##################################################')
print('WARNING: This code is has never been used for real')
print('##################################################')

class WorkerThread(threading.Thread):
    def __init__(self, query_queue, ignored_events):
        super(WorkerThread, self).__init__()
        self._client = bigquery.Client()
        self._query_queue = query_queue
        self._ignored_events = ignored_events
        self._event_types = defaultdict(int)
        self._page_table = TableTracker("page", 1000)
        self._tracks_table = TableTracker("tracks", 1000, is_tracks=False)
        self._other_count = defaultdict(int)
        self._event_tables = {}
        self._track_count = 0
        self._total_time = 0
        self._files = 0
        self._ignored_event_count = 0
        #self._start_time = time.mktime(args.start_date.utctimetuple()) if args.start_date else 0
        #self._end_time = time.mktime(args.end_date.utctimetuple()) if args.end_date else sys.maxint

        print("Ignored events: %s" % str(self._ignored_events))

    def total_count(self):
        return self._tracks_table._total_count

    def run(self):
        ended = False
        while not ended:
            item = self._query_queue.get()
            if not item:
                ended = True
            else:
                self.process_query(item)
            self._query_queue.task_done()

    def process_query(self, query):
        start_time = time.time()

        query_results = self._client.run_sync_query(query)
        query_results.use_legacy_sql = False
        query_results.use_query_cache = False

        query_results.run()

        schema = query_results.schema

        # Drain the query results by requesting a page at a time.
        page_token = None

        while True:
            rows, total_rows, page_token = query_results.fetch_data(
                max_results=100,
                page_token=page_token)

            for row in rows:
                self.process_row(row, schema)

            if not page_token:
                break
        self._total_time += (time.time() - start_time)
        self._files += 1

    def process_row(self, row, schema):
        """Process a top-level row of the table"""
        self._has_userid = "userId" in [field.name for field in schema]
        if args.require_user_id and not self._has_userid:
            return
        # Fake adding an entire row to increment Table's _total_count
        self._tracks_table.add_entry({}, self._has_userid)
        for index, value in enumerate(row):
            field = schema[index]
            self.process_column(value, field, "")

    def process_column(self, value, schema, column_name):
        """Process a top-level column of a row in the table"""
        if schema.mode == 'REPEATED':
            for index, value in enumerate(value):
                if schema.field_type == 'RECORD':
                    self.process_record(schema.name, value, schema.fields, column_name)
                else:
                    value_name = _column_name_join(column_name, schema.name)
                    self.analyze_value(value_name, schema.field_type, value)
        elif schema.field_type == 'RECORD':
            self.process_record(schema.name, value, schema.fields, column_name)
        else:
            value_name = _column_name_join(column_name, schema.name)
            self.analyze_value(value_name, schema.field_type, value)

    def process_record(self, name, value, schema, column_name):
        record_name = _column_name_join(column_name, name)
        schema_map={ field.name: field for field in schema }
        if value is not None:
            for k, v in list(value.items()):
                self.process_column(v, schema_map[k], record_name)

    def analyze_value(self, value_name, type, value):
        self._tracks_table._add_column(value_name, value)


def _column_name_join(left, right):
    return right if len(left) == 0 else ".".join([left, right])

def create_table_info(schema, table_tracker):
    table_info = TableInfo(schema, table_tracker._name)
    table_info.set_rows(table_tracker._total_count)
    #table_info.add_column(ColumnInfo("userId", "varchar"))
    #table_info.add_column(ColumnInfo("receivedAt", "timestamp"))

    for column_tracker in list(table_tracker._columns.values()):
        if column_tracker._datatype is None:
            continue

        name = column_tracker._name
        null_fraction = 1.0 - float(column_tracker._total_non_null_count) / table_tracker._total_count
        column = ColumnInfo(name, column_tracker._datatype)
        column.null_fraction = null_fraction

        top_values = column_tracker.get_top_values(0.00001)
        if len(top_values) > 0 and top_values[0] > 100:
            for value in top_values:
                column.add_group(value[0], value[1] / 100.0)

        table_info.add_column(column)
    return table_info


parser = argparse.ArgumentParser(description='Analyzes Segment JSON data in S3')
#parser.add_argument('--s3_src', type=str, help='s3 path of data ', nargs = '+', required=True)
#parser.add_argument('--s3_profile', type=str, help='s3 profile to use for data access', default = "s3_data_user")
parser.add_argument('--sample_ratio', type=float, help='fraction of data to read', default = 0.01)
#parser.add_argument('--start_date', type = segment_utils.as_date, help = 'Start date of data to analyze', default = None)
#parser.add_argument('--end_date', type = segment_utils.as_date, help = 'End date of data to analyze', default = None)
parser.add_argument('--sampling_seed', type = int, help = 'Random seed for sampling', default = 91876)
parser.add_argument('--max_files', type = int, help = 'Maximum files to sample', default = None)
parser.add_argument('--schema_info', type = str, help = 'SchemaInfo file to create', required = True)
#parser.add_argument('--filter_config', type=str, help='JSON filter config file', required=True)
parser.add_argument('--client', type = str, help = 'Client name', required = True)
parser.add_argument('--require_user_id', type = bool, help = "If true, then only count rows with userId set", default = False)

args = parser.parse_args()
rng = Random()
rng.seed(args.sampling_seed)

#filter_config = SchemaFilter.load_from_json(args.filter_config)

query_queue = queue.Queue(20)

thread = WorkerThread(query_queue, [])
thread.daemon = True
thread.start()
file_count = 0
start_time = time.time()
       
query_queue.put('SELECT * FROM `analytics-api-165801.138255571.ga_sessions_20170523` where totals.transactions > 0 limit 2')
query_queue.put(False)
query_queue.join()
print("Other types: %s" % str(thread._other_count))

print(thread._page_table.print_pretty(0.02))
print(thread._tracks_table.print_pretty(0.02))

# for table in sorted(thread._event_tables.values(), key = lambda (t): t._total_count, reverse = True):
#       if table._total_count > thread.total_count() * args.min_event_fraction:
#               print table.print_pretty(0.02)
#       else:
#               print "Skipping %s.  Too small %5.4f%%" % (table._name, table._total_count * 100.0 / thread.total_count())

track_info = create_table_info(args.client, thread._tracks_table)
page_info = create_table_info(args.client, thread._page_table)

# First pass, take filter min and use threshold 20x lower
#min_event_fraction = filter_config.min_join_table_row_fraction / 20
join_tables = {}
#for table_tracker in thread._event_tables.values():
#    if table_tracker._total_count > thread.total_count() * min_event_fraction:
#        join_tables[table_tracker._name] = create_table_info(args.client, table_tracker)

schema_info = SchemaInfo(args.client, track_info, page_info, join_tables)

save_schema(args.schema_info, schema_info)



from bytegain.custom.model_data.row_handler import RowHandler
from bytegain.custom.model_data.bucket import ValueBucket
from bytegain.custom.model_data.extractor import RowExtractor
from bytegain.custom.cust.apartmentlist.al_multi_row_handler import ALMultiRowHandler


# # This WHERE is just for creating features.  For training data, see leads_from_redshift.
# WHERE =  """
# interest_created_at > timestamp '2016-01-01' and
# interest_created_at < timestamp '2016-08-01'
# """
# def _get_selector():
#       return read_from_db.Selector("apartmentlist.bytegain_leads", WHERE, "leased_at_notified_property", "interest_id")

"""
Creates a RowHandler for parsing AL data and handing to XGBoost.
"""


def create_multi_row_handler(feature_json):
    return ALMultiRowHandler(feature_json, label_bucket = ValueBucket("leased", RowExtractor("leased_at_notified_property"), has_null = False),
            id_extractor= RowExtractor("interest_id"))



import bytegain.custom.cust.apartmentlist.al_xgb
from bytegain.custom.cust.apartmentlist.feature_filter import FeatureFilter

from .common import EXCLUDED_FEATURES

NUMERICAL_FEATURES = [
    'preferences_lease_length',
    'preferences_move_urgency',
]

CATEGORY_FEATURES = [
    'property_city_id',
]

BUCKET_FEATURES = [
    'registered_source_app',
]

VERSION = '1.2.2'

filter = FeatureFilter(EXCLUDED_FEATURES, NUMERICAL_FEATURES, CATEGORY_FEATURES, [])

config = {'table': 'ml.finance_20190724',
          'model_name': 'al_finance',
          'model_version': VERSION,
          'feature_json': 'al_finance_v%s' % VERSION,
          'filter': filter,
          'outcome_field': 'billable_lease_at_property',
          'control_field': 'lsd_control',
          'control_weight': 1,
          'id_field': 'event_id',
          'random_seed': 91876,
          'max_rounds': 1001,
          'max_depth': 2,
          'train_step_size': 0.35,
          'min_child_weight': 1000,
          'l2_norm': 200,
          'downsample': True}



prod_config = dict(config)
prod_config['downsample'] = False



"""
Reads rows from a cursor and converts into a dataset using buckets.
"""
import pyximport

pyximport.install()

import datetime
import itertools
import os
import time
import numpy
import psycopg2
import psycopg2.extras
import psycopg2.extensions
from hashlib import md5

import snowflake.connector
from snowflake.connector import DictCursor

import bytegain.custom.model_data.dataset as dataset
import bytegain.custom.model_data.row_handler as row_handler_module

import bytegain.custom.model_data.row_handler_proto.row_handler_pb2 as row_handler_pb2
from bytegain.custom.model_data.row_handler_proto.row_handler_pb2 import Extractor


def timestamp_to_epoch(value, curs, as_int=True):
    if value == None:
        return None
    # Sometimes these come in with fractional seconds, sometimes not.
    if '.' in value:
        dt = datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S.%f')
    else:
        dt = datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
    return int(time.mktime(dt.utctimetuple())) if as_int else dt


# Use floats instead of DECIMAL since floats are way faster
DEC2FLOAT = psycopg2.extensions.new_type(
    psycopg2.extensions.DECIMAL.values,
    'DEC2FLOAT',
    lambda value, curs: float(value) if value is not None else None)
psycopg2.extensions.register_type(DEC2FLOAT)

TIMESTAMP2EPOCH = psycopg2.extensions.new_type(
    psycopg2.DATETIME.values,
    'TIMESTAMP2EPOCH',
    timestamp_to_epoch)
psycopg2.extensions.register_type(TIMESTAMP2EPOCH)


def dates_as_timestamp(as_timestamp=True):
    TIMESTAMP2EPOCH = psycopg2.extensions.new_type(
        psycopg2.DATETIME.values,
        'TIMESTAMP2EPOCH',
        lambda value, curs: timestamp_to_epoch(value, curs, not as_timestamp))
    psycopg2.extensions.register_type(TIMESTAMP2EPOCH)


class ModSelector(object):
    def __init__(self, mods, test_mod, seed=1763):
        self._mods = mods
        self._test_mod = test_mod
        self._seed = seed

    def get_order_by(self, field):
        # return "order by sort_mod, interest_id"
        return "order by ((((%s + %d) * 7) * 17) * 103) %% 9999, %s" % (field, self._seed, field)

    def get_where(self, field, is_test):
        return "((((%s + %d) * 23) * 19) * 97) %% %d %s %d " % (
        field, self._seed, self._mods, "=" if is_test else "!=", self._test_mod)


class Selector(object):
    def __init__(self, table, where, label_field, id_field, secondary_sort=None, positive_outcome_rate_field = None):
        self._table = table
        self._label_field = label_field
        self._id_field = id_field
        self._secondary_sort = secondary_sort

        if positive_outcome_rate_field:
            term = "({label} or CORE.public.f_bucket_number({id}, 1000) < ({rate_field} * 1000))".format(
                label=self._label_field,
                id = self._id_field,
                rate_field = positive_outcome_rate_field)
            if where:
                where = where + " AND " + term
            else:
                where = term

        self._where = where

    # Returns the where statement, or "" if there is no where select
    def get_where(self):
        return ("where %s" % self._where) if self._where is not None else ""

    def create_statement(self, select):
        statement = "select %s, %s, %s from %s %s order by %s" % (
            self._id_field, self._label_field, select, self._table, self.get_where(),
            self._id_field if self._secondary_sort is None else ("%s, %s" % (self._id_field, self._secondary_sort)))
        # print statement
        return statement

    # def create_statement_for_queue(self, select, mod_selector, is_test):
    #     where = self._where
    #     if where is not None:
    #         where += " AND "
    #     where += mod_selector.get_where(self._id_field, is_test)
    #
    #     statement = "select %s, %s, %s from %s where %s %s" % (
    #         self._id_field, self._label_field, select, self._table, where, mod_selector.get_order_by(self._id_field))
    #     return statement


def create_connection(warehouse=None):
    return snowflake.connector.connect(
        account="apartmentlist.us-east-1",
        user=os.environ["SNOWFLAKE_USER"],
        password=os.environ["SNOWFLAKE_PASSWORD"],
        warehouse=warehouse
    )

def generate_row_handler_stats(selector, row_handler):
    start_time = time.time()
    with create_connection() as connection:
        _extract_auto_params(connection, selector, row_handler)
    print("RowHandler stat time: %4.2fs" % (time.time() - start_time))


class IterableDataset(object):
    def __init__(self, statement, connection, row_handler, one_epoch=False):
        self._one_epoch = one_epoch
        self._row_handler = row_handler
        self._statement = statement
        self._connection = connection
        self._cursor = connection.cursor()
        self._initialize_cursor(True)

    def _initialize_cursor(self, update_row_handler=False):
        start_time = time.time()
        self._cursor.execute(self._statement)
        print("execute cursor time: %5.1fs" % (time.time() - start_time))
        if update_row_handler:
            self._header = [x.name for x in self._cursor.description]
            for bucket in self._row_handler._buckets:
                extractor = bucket._extractor
                extractor.update_fields_by_id(self._header)

    def __iter__(self):
        return self

    def __next__(self):
        row = self._cursor.fetchone()
        if row is None:
            if self._one_epoch:
                raise StopIteration
            else:
                self._initialize_cursor()
                row = self._cursor.fetchone()
                if row is None:
                    # This happens only with empty query
                    raise StopIteration

        data = self._row_handler.get_data_from_row(row)

        zipped_row = dict(list(zip(self._header, row)))
        label = self._row_handler.get_label_from_row(zipped_row)
        id = self._row_handler.get_id_from_row(zipped_row)

        return data, label, id

    def __del__(self):
        self._connection.close()


def create_enqueue_operator(password, dbname, selector, row_handler, mod_selector, is_test=False):
    connection = create_connection(password, dbname)
    row_handler_module.set_fast_mode(True)
    start_time = time.time()
    statement = selector.create_statement_for_queue(get_select(row_handler), mod_selector, is_test)

    # For testing, we only iterate once through
    return IterableDataset(statement, connection, row_handler, one_epoch=is_test)


"""
if password is None, then password is read from ENV.
For instance if dbname=apartmentlist, then looks for ENV variable: APARTMENTLIST_PASS
"""


def create_datasets(selector, row_handler, test_fraction=0.10, row_limit=None, shuffle_rows = True):
    dates_as_timestamp(False)
    row_handler_module.set_fast_mode(True)
    categorical_indicies = row_handler.get_categorical_indicies()
    start_time = time.time()
    with create_connection() as connection:
        connection.autocommit = False
        # By default psycopg2 uses a client-side cursor which downloads all data before returning a single row
        with connection.cursor() as cursor:
            cursor.itersize = 100000
            statement = selector.create_statement(get_select(row_handler))
            if row_limit:
                statement = statement + (" limit %d" % row_limit)
            print("Statement: %s" % statement)
            cursor.execute(statement)
            cursor.fetchone()
            row_handler.replace_field_names_with_index(cursor)
            datasets = _create_datasets_from_cursor(cursor, row_handler, test_fraction, shuffle_rows)
    datasets.categorical_indicies = categorical_indicies
    print("Load time: %4.2fs" % (time.time() - start_time))
    return datasets


def get_select(row_handler):
    selects = []
    for bucket in row_handler.buckets:
        extractor = bucket._extractor
        selects.append(apply_clamp(extractor))
        field2 = apply_clamp(extractor, field2=True)
        if field2:
            selects.append(field2)
    if row_handler._control_extractor:
        selects.extend(row_handler._control_extractor.get_fields())
    selects.extend(row_handler.extra_fields())
    return ",".join(selects)


def _create_datasets_from_cursor_multirow(cursor, multi_row_handler, test_fraction):
    print ("Creating multi-row")
    data = []
    labels = []
    ids = []
    count = 0
    start_time = time.time()
    for row in cursor:
        count += 1
        entry = multi_row_handler.process_row(row)
        if entry is not None:
            data.append(entry[0])
            labels.append(entry[1])
            ids.append(entry[2])

        if (count % 100000) == 0:
            print("%d -> %d" % (count, len(data)))

    print("rows: %d in %ds" % (len(data), time.time() - start_time))
    # data = numpy.array(data, dtype=numpy.float)
    labels = numpy.array(labels, dtype=numpy.float)
    ids = numpy.array(ids)
    # row_handler.print_summary()
    return data, labels, ids, multi_row_handler
    # return dataset.DataSets(multi_row_handler, data, labels, ids, test_fraction)


def _create_datasets_from_cursor(cursor, row_handler, test_fraction, shuffle_rows):
    data = []
    labels = []
    ids = []
    weights = []
    is_control = []
    for row in cursor:
        data.append(row_handler.get_data_from_row(row))
        labels.append(row_handler.get_label_from_row(row)[0])
        ids.append(row_handler.get_id_from_row(row))
        weights.append(row_handler.get_weight_from_row(row))
        is_control.append(row_handler.is_control_from_row(row))

    print("rows: %d" % len(data))
    data = numpy.array(data, dtype=numpy.float)
    labels = numpy.array(labels, dtype=numpy.float)
    ids = numpy.array(ids)
    weights = numpy.array(weights)
    is_control = numpy.array(is_control)
    row_handler.print_summary()
    return dataset.DataSets(row_handler, data, labels, ids, weights, test_fraction, shuffle_rows, is_control=is_control)


def apply_clamp(extractor, with_name=True, field2=False):
    if field2:
        field = extractor.get_field2()
        if field is None:
            # No field2, so do nothing
            return None
    else:
        field = extractor._field_name
    field_select = field
    if hasattr(extractor, "has_min") and extractor.has_min():
        field_select = "GREATEST(%s, %f)" % (field, extractor._min_val)
    if hasattr(extractor, "has_max") and extractor.has_max():
        field_select = "LEAST(%s, %f)" % (field_select, extractor._max_val)
    if field_select != field and with_name:
        field_select = "(case when %s is not null then %s else null end) as %s" % (field, field_select, field)
    return field_select


def _extract_auto_params(connection, selector, row_handler):
    # Pull out percentiles
    where = (selector._where + " AND") if selector._where else ""
    for bucket in row_handler.buckets:
        extractor = bucket._extractor
        if extractor._proto_type == Extractor.PERCENTILE_NORM:
            with connection.cursor() as cursor:
                # TODO(chris): Figure out a better abstraction for using selector._where.
                s = "select max(value), percentile from (select %s as value, ntile(100) over (order by %s asc) as percentile from bytegain_leads where %s %s is not null) group by 2 order by 2;" % (
                extractor._field, extractor._field, where, extractor._field)
                cursor.execute(s)
                percentiles = [float(x[0]) for x in cursor.fetchall()][0:-1]
                extractor.set_auto_values(percentiles)

    # Pull out categories for CategoryBucket
    for bucket in row_handler._buckets:
        if hasattr(bucket, "set_categories"):
            with connection.cursor() as cursor:
                s = "select %s, count(1) from %s where %s %s is not null group by 1 order by 2 desc" % (
                    bucket._extractor._field_name, selector._table, where, bucket._extractor._field_name)
                cursor.execute(s)
                categories = [{"name": x[0], "count": x[1]} for x in cursor.fetchall()]
                bucket.set_categories(categories)

    for bucket in row_handler._buckets:
        with connection.cursor() as cursor:
            s = "select %s from %s where %s %s is not null limit 1" % (
                bucket._extractor._field_name, selector._table, where, bucket._extractor._field_name)
            cursor.execute(s)
            # TODO(jjs): If all values are None?
            value = cursor.fetchone()[0]
            if isinstance(value, float):
                datatype = row_handler_pb2.Extractor.FLOAT
            elif isinstance(value, str):
                datatype = row_handler_pb2.Extractor.STRING
            elif isinstance(value, bool):
                datatype = row_handler_pb2.Extractor.BOOL
            elif isinstance(value, int) or isinstance(value, int):
                datatype = row_handler_pb2.Extractor.INT
            else:
                raise Exception("Unknown datatype: %s for %s" % (type(value), bucket._extractor._field_name))
            bucket._extractor._data_type = datatype

    select = ""
    for bucket in row_handler.buckets:
        extractor = bucket._extractor
        if hasattr(extractor, "_field"):
            field = extractor._field_name
            field_select = apply_clamp(extractor, with_name=False)
            select += "avg(case when %s is not null then 1.0 else 0.0 end) as %s, " % (field_select, field + "_null")
            if hasattr(extractor, "_sigma") and extractor._data_type is not row_handler_pb2.Extractor.STRING:
                if extractor._data_type == row_handler_pb2.Extractor.BOOL:
                    cast = "int::float"
                else:
                    cast = "float"
                select += ("stddev(%s::%s) as %s, avg(%s::%s) as %s, " % (
                field_select, cast, field + "_std", field_select, cast, field + "_avg"))
    if len(select) > 0:
        select = select[0:-2]
        full_statement = "select %s from %s %s" % (
            select, selector._table, ("where %s" % selector._where) if selector._where is not None else "")
        # print full_statement
        with connection.cursor(DictCursor) as cursor:
            cursor.execute(full_statement)
            row = cursor.fetchone()
            for bucket in row_handler.buckets:
                extractor = bucket._extractor
                if hasattr(extractor, "_field_name"):
                    field = bucket._extractor._field_name
                    extractor.set_null_fraction(row[field.upper() + "_NULL"])
                    if hasattr(extractor,
                               "set_auto_values") and extractor._data_type is not row_handler_pb2.Extractor.STRING:
                        std = row[field.upper() + "_STD"]
                        mean = row[field.upper() + "_AVG"]
                        bucket._extractor.set_auto_values(mean, std)



import datetime
import logging

from bytegain.custom.model_data.model_config import get_epoch_times_for_rows
from bytegain.zap.prediction_instance import PredictionInstance, increment_counter

MINUTE = datetime.timedelta(minutes=1)

def row_is_modal_view(row):
    return 'event' in row and row['event'] == 'modals' and row['name'] == 'signup' and row['action'] == 'view'


def row_is_sign_up(row):
    return 'event' in row and row['event'] in ('user_signed_up', 'social_sign_in')

def row_is_logged_in(row):
    return 'event' in row and row['event'] in ('user_logged_in', 'user_social_login')

def row_is_interest(row):
    return 'event' in row and row['event'] == 'cycle_choice'

def row_is_positive_interest(row):
    return row_is_interest(row) and row['action'] in ('maybe', 'love-it', 'email', 'go-visit')

def row_is_notification(row):
    return 'event' in row and row['event'] == 'interest' and (
           'leased_at_notified_property' in row and row['leased_at_notified_property'] in ('t', 'f'))


class LoginInstanceGenerator(object):
    """ Instance generator for Apartmentlist's signup modal.

    Positive instances take one of two forms:
    1) modals name=signup action=view -> modals name=signup action=submit -> user_signed_up
    2) modals name=signup action=view -> modals name=signup action=email_taken -> social_sign_in

    """
    def get_instances(self, rows):
        epoch_times = get_epoch_times_for_rows(rows)

        # Generate a prediction event for each login modal view.
        instances = []
        for trigger_i, trigger_row in enumerate(rows):
            # Don't generate any instances after the user's first login.
            if row_is_logged_in(trigger_row):
                increment_counter('users_with_login')
                break

            if row_is_modal_view(trigger_row):
                # Label is whether they logged in in the next 10 minutes.
                logged_in = False
                for i in range(trigger_i + 1, len(rows)):
                    time_since_trigger = epoch_times[i] - epoch_times[trigger_i]
                    if time_since_trigger > 10*MINUTE:
                        break
                    if row_is_sign_up(rows[i]):
                        increment_counter('signin-%s' % rows[i]['event'])
                        logged_in = True
                        break
                instances.append(PredictionInstance(0, trigger_i, logged_in))
                break

        return instances


class ModalBounceInstanceGenerator(object):
    """ Instance generator whether showing Apartmentlist's signup modal will cause a bounce.

    Negative instances take one of three forms:
    1) modals name=signup action=view -> modals name=signup action=submit -> user_signed_up
    2) modals name=signup action=view -> modals name=signup action=email_taken -> social_sign_in
    3) Still on site after 5min

    Positive instances look like this:
    1) modals name=signup action=view -> everything else

    """
    def get_instances(self, rows):
        epoch_times = get_epoch_times_for_rows(rows)

        # Generate a prediction event for each login modal view.
        instances = []
        for trigger_i, trigger_row in enumerate(rows):
            # Don't generate any instances after the user's first login.
            if row_is_logged_in(trigger_row):
                increment_counter('users_with_login')
                break

            if row_is_modal_view(trigger_row):
                bounce = True
                for i in range(trigger_i + 1, len(rows)):
                    if row_is_sign_up(rows[i]):
                        increment_counter('xbounce-signup')
                        bounce = False
                        break
                    seconds_since_trigger = epoch_times[i] - epoch_times[trigger_i]
                    if seconds_since_trigger > 5*MINUTE and seconds_since_trigger < 60*MINUTE:
                        increment_counter('xbounce-activity')
                        bounce = False
                        break

                instances.append(PredictionInstance(0, trigger_i, bounce))
                break

        return instances

class AbandonInstanceGenerator(object):
    """ Instance generator for Apartmentlist's cycling matches UI.

    Positive instances are where user continued to look at more matches

    """
    def get_instances(self, rows):
        positive = False
        instances = []
        # Find the last positive interest
        last = 0
        interests = 0
        for i, row in enumerate(rows):
            if row_is_positive_interest(row):
                interests += 1
                last = min(i + 1, len(rows) - 1)

        total_notifications = AbandonInstanceGenerator.count_notifications(rows[0:last + 1])
        increment_counter("Notifications for user %2d" % total_notifications)
        increment_counter("Total positive interests", interests)
        increment_counter("Total notifications", total_notifications)

        for trigger_i, trigger_row in enumerate(rows):
            if not positive or row_is_positive_interest(trigger_row):
                positive = True

            if positive and row_is_positive_interest(trigger_row):
                notifications = AbandonInstanceGenerator.count_notifications(rows[trigger_i + 1: last + 1])
                future_data = "%d %d" % (total_notifications, notifications)
                increment_counter("Future_Data: %s" % future_data)
                outcome = False
                for i in range(trigger_i + 1, len(rows)):
                    if row_is_positive_interest(rows[i]):

                        instances.append(PredictionInstance(
                            0, trigger_i + 1, True, i, future_data =future_data))
                        outcome = True
                        break
                if not outcome:
                    instances.append(PredictionInstance(
                        0, trigger_i + 1, False, future_data =future_data))
        return instances

    @staticmethod
    def count_notifications(rows):
        count = 0
        for row in rows:
            if row_is_notification(row):
                count +=1
        return count
def filter_view_button(event):
    return 'event' not in event or event['event'] != 'buttons' or event['action'] != 'view'